// FileFinder.cpp
//
// This program recurses through all directories and logs
// file information to note any changes.
//

#include "Main.h"

DWORD attribCnt[32];	// 32 bits to check

FILENODE_DB *getFileNode(WIN32_FIND_DATA *fileData)
{
	FILENODE_DB *ptrFN;
	SYSTEMTIME modified;
	FILETIME mTime;

	if(gFileNodeDB_Count >= gMaxFileNodeDB_Count)
	{
		fprintf(stderr, "Warning! Ran out of file node allocations - using malloc()\n\n");
		ptrFN = (FILENODE_DB *) malloc(sizeof(FILENODE_DB));
		if(ptrFN == NULL)
		{
			fprintf(stderr, "Memory Error! Could not allocate %u bytes for a file node.\n\n", sizeof(FILENODE_DB));
			return NULL;
		}
	}
	else
	{
		ptrFN = &gptrTopFileNodeDB[gFileNodeDB_Count];
		gFileNodeDB_Count++;
	}

	ptrFN->fileNameIndex = 0;	// may not need this
	strcpy(ptrFN->filePathName, gCurrentPath);
	strcat(ptrFN->filePathName, fileData->cFileName);

	ptrFN->fileSizeQ.HighPart = fileData->nFileSizeHigh;
	ptrFN->fileSizeQ.LowPart = fileData->nFileSizeLow;

	memset(&modified, 0, sizeof(SYSTEMTIME));
	FileTimeToLocalFileTime(&fileData->ftLastWriteTime, &mTime);
	FileTimeToSystemTime(&mTime, &modified);

	ptrFN->year = modified.wYear;
	ptrFN->month = modified.wMonth;
	ptrFN->day = modified.wDay;
	ptrFN->hour = modified.wHour;
	ptrFN->minute = modified.wMinute;
	ptrFN->second = modified.wSecond;
	ptrFN->milli = modified.wMilliseconds;

	ptrFN->nextFile = NULL;

	return ptrFN;
} // getFileNode

// for each file, record which attributes
void countAttribs(DWORD attrib)
{
	int i;
	DWORD mask;

	mask = 1;
	for(i = 0; i < 32; i++)
	{
		if( (attrib & mask) != 0) attribCnt[i]++;
		mask *= 2;
	}
	return;
} // countAttribs

void printAttribs(FILE *fptr)
{
	DWORD i, mask, tmpD;

	mask = 1;
	for(i = 0; i < 32; i++)
	{
		switch(mask)
		{
		case FILE_ATTRIBUTE_READONLY:
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_READONLY = %u\n", mask, attribCnt[i]);
			break;

		case FILE_ATTRIBUTE_HIDDEN:
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_HIDDEN = %u\n", mask, attribCnt[i]);
			break;

		case FILE_ATTRIBUTE_SYSTEM:
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_SYSTEM = %u\n", mask, attribCnt[i]);
			break;

		case FILE_ATTRIBUTE_DIRECTORY:
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_DIRECTORY = %u\n", mask, attribCnt[i]);
			break;

		case FILE_ATTRIBUTE_ARCHIVE:
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_ARCHIVE = %u\n", mask, attribCnt[i]);
			break;

		case FILE_ATTRIBUTE_DEVICE:
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_DEVICE = %u\n", mask, attribCnt[i]);
			break;

		case FILE_ATTRIBUTE_NORMAL:
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_NORMAL = %u\n", mask, attribCnt[i]);
			break;

		case FILE_ATTRIBUTE_TEMPORARY:	// 0x100
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_TEMPORARY = %u\n", mask, attribCnt[i]);
			break;

		case FILE_ATTRIBUTE_SPARSE_FILE:
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_SPARSE_FILE = %u\n", mask, attribCnt[i]);
			break;

		case FILE_ATTRIBUTE_REPARSE_POINT:
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_REPARSE_POINT = %u\n", mask, attribCnt[i]);
			break;

		case FILE_ATTRIBUTE_COMPRESSED:
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_COMPRESSED = %u\n", mask, attribCnt[i]);
			break;

		case FILE_ATTRIBUTE_OFFLINE:	// 0x1000
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_OFFLINE = %u\n", mask, attribCnt[i]);
			break;

		case FILE_ATTRIBUTE_NOT_CONTENT_INDEXED:
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_NOT_CONTENT_INDEXED = %u\n", mask, attribCnt[i]);
			break;

		case FILE_ATTRIBUTE_ENCRYPTED:
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_ENCRYPTED = %u\n", mask, attribCnt[i]);
			break;

		case FILE_ATTRIBUTE_VIRTUAL:
			fprintf(fptr, "%08x] FILE_ATTRIBUTE_VIRTUAL = %u\n", mask, attribCnt[i]);
			break;

		default:
			fprintf(fptr, "%08x] DEFAULT = %u\n", mask, attribCnt[i]);
			break;
		} // END SWITCH
		mask = mask * 2;
	} // end for loop
	return;
} // printAttribs

// new file node
FILENODE *newFileNode(FILENODE *fn)
{
	FILENODE *pf = NULL;
	size_t x;

	pf = (FILENODE *) malloc(sizeof(FILENODE));

	if(pf == NULL)
	{
		printf("Could not allocate memory for FILENODE\n\n");
		return(NULL);
	}

	pf->fileName = (char *) malloc(strlen(fn->fileName)+1);
	if(pf->fileName == NULL)
	{
		printf("Could not allocate memory for filename %s\n\n", fn->fileName);
		free(pf);
		return(NULL);
	}

	strcpy(pf->fileName, fn->fileName);
	x = strlen(pf->fileName);
	x--;
	if( *(pf->fileName+x) == 10) *(pf->fileName+x) = 0;

	strcpy(pf->created, fn->created);
	strcpy(pf->accessed, fn->accessed);
	strcpy(pf->modified, fn->modified);
	//TIME
	strcpy(pf->createdTime, fn->createdTime);
	strcpy(pf->accessedTime, fn->accessedTime);
	strcpy(pf->modifiedTime, fn->modifiedTime);
	// TIME
	pf->fileSizeQ.QuadPart = fn->fileSizeQ.QuadPart;
	pf->attrib = fn->attrib;

	countAttribs(fn->attrib);

	pf->checked = 0;	// used for search
	pf->next = NULL;
	return(pf);
} // newFileNode

// allocate memory for a directory node
DIRNODE *newDirNode(char *dirName)
{
	DIRNODE *pd = NULL;
	size_t x;

	pd = (DIRNODE *) malloc(sizeof(DIRNODE));

	if(pd == NULL)
	{
		printf("Could not allocate memory for DIRNODE.\n\n");
		return(NULL);
	}

	pd->dirName = (char *) malloc(strlen(dirName)+1);
	if(pd->dirName == NULL)
	{
		printf("Could not allocate memory for directory name %s\n\n", dirName);
		free(pd);
		return(NULL);
	}

	strcpy(pd->dirName, dirName);

	// strip line feed character
	x = strlen(pd->dirName);
	x--;
	if(*(pd->dirName+x) == 10) *(pd->dirName+x) = 0;

	pd->checked = 0;	// used in search
	pd->subDirList = NULL;
	pd->fileList = NULL;
	pd->next = NULL;
	return(pd);
} // newDirNode

// show the directory tree structure
// ASSUMES at level zero no other directories
// so if pass lower directory, will only print sub structure
void printDirTree(FILE *fptr, DIRNODE *topDir, bool showFiles)
{
	int level, i;
	char curDir[MAX_PATH], space[1024];
	DIRNODE *pDir[256];
	FILENODE *pf;
	size_t x;

	memset(space, ' ', 1024);
	level = 1;
	pDir[level] = topDir;
	strcpy(curDir, pDir[level]->dirName);

	while(level > 0)
	{
		fprintf(fptr, "\n ... %s\n", curDir);

		// print file list
		space[level * 2] = 0;
		if(showFiles)
		{
			pf = pDir[level]->fileList;
			while(pf != NULL)
			{
				fprintf(fptr, "%s  --> %s\n", space, pf->fileName);
				pf = pf->next;
			}
			fflush(fptr);
		}

		if(pDir[level]->subDirList != NULL) // go down level
		{
			space[level * 2] = ' ';
			level++;
			space[level * 2] = 0;
			pDir[level] = pDir[level-1]->subDirList;
			strcat(curDir, pDir[level]->dirName);
			continue;
		}

Again:
		if(pDir[level]->next != NULL)
		{
			x = strlen(curDir) - 1;
			while(x >= 0)
			{
				if(curDir[x] == '\\')
				{
					curDir[x] = 0;
					break;
				}
				curDir[x] = 0;
				x--;
			}

			// remove a dir
			pDir[level] = pDir[level]->next;
			strcat(curDir, pDir[level]->dirName);
			continue;
		}

		// go up level
		x = strlen(curDir) - 1;
		while(x >= 0)
		{
			if(curDir[x] == '\\')
			{
				curDir[x] = 0;
				break;
			}
			curDir[x] = 0;
			x--;
		}
		space[level * 2] = ' ';
		level--;
		space[level * 2] = 0;
		if(level > 0) goto Again;
	} // while

	return;
} // printDirTree

// convert a hex to a DWORD
DWORD hex2DWORD(char *data, int len)
{
	DWORD result;
	int i, factor;
	char c;

	i = len-1;
	result = 0;
	factor = 1;

	while(i >= 0)
	{
		c = *(data+i);

		if(c >= 48 && c <= 57) // number
		{
			c = c - 48;
		}
		else if(c >= 'A' && c <= 'F')
		{
			c = c - 0x41 + 10;
		}
		else if(c >= 'a' && c <= 'f')
		{
			c = c - 0x61 + 10;
		}
		else
		{	printf("Error, illegal hex value - %02x, i=%d.\n\n", c, i);
			return(0);
		}
		result = result + factor * c;
		factor = factor * 16;
		i--;
	} // while
	return(result);
} // hex2DWORD

// convert a hex to a DWORD
DWORD hex2QWORD(char *data, int len)
{
	QWORD result;
	int i, factor;
	char c;

	i = len-1;
	result = 0;
	factor = 1;

	while(i >= 0)
	{
		c = *(data+i);

		if(c >= 48 && c <= 57) // number
		{
			c = c - 48;
		}
		else if(c >= 'A' && c <= 'F')
		{
			c = c - 0x41 + 10;
		}
		else if(c >= 'a' && c <= 'f')
		{
			c = c - 0x61 + 10;
		}
		else if(c == 'x' || c == 'X')
		{
			break;	// done
		}
		else
		{	printf("Error, illegal hex value - %02x, i=%d.\n\n", c, i);
			return(0);
		}
		result = result + factor * c;
		factor = factor * 16;
		i--;
	} // while
	return(result);
} // hex2QWORD


// find one directory in lost of other
DIRNODE * findDir(DIRNODE *d1, DIRNODE *d2)
{
	while(d2 != NULL)
	{
		if( strcmp(d1->dirName, d2->dirName) == 0 )
		{
			d1->checked = 1;
			d2->checked = 1;
			return(d2);
		}
		d2 = d2->next;
	}
	return(NULL);
} // findDir

// find a file in the list
FILENODE *findFileInList(FILENODE *f1, FILENODE *f2)
{
	while(f2 != NULL)
	{
		if(strcmp(f1->fileName, f2->fileName) == 0) // name match
		{
			f1->checked = 1;
			f2->checked = 1;
			return(f2);
		}
		f2 = f2->next;
	}
	return(NULL);
} // findFileInList

// need a 64 bit absolute value function
QWORD absval(__int64 number)
{
	if(number >= 0) return number;

	return( number * -1 );
} // abs

// checks two files' attributes
void matchFileAttributes(FILENODE *f1, FILENODE *f2, char *dirName1, char *dirName2)
{
	int flag = 0;

	if(f1->fileSizeQ.QuadPart != f2->fileSizeQ.QuadPart)
	{
		fprintf(gFptr, "\n********************************\n  <5> FILE PROPERTY MISMATCH:  %s\n\n", dirName1);
		fprintf(gFptr, "      FILE <%s>\n", f1->fileName);
		flag = 1;


		fprintf(gFptr, "\n        <6>  - FILE SIZE CHANGE OF <%lld> BYTES\n", absval( (__int64) (f1->fileSizeQ.QuadPart - f2->fileSizeQ.QuadPart) ));
		fprintf(gFptr, "      file1:%lld\n      file2:%lld\n", f1->fileSizeQ.QuadPart, f2->fileSizeQ.QuadPart);
	}
//*
	if(f1->attrib != f2->attrib)
	{
		if(flag == 0)
		{
			fprintf(gFptr, "\n********************************\n  <7> FILE PROPERTY MISMATCH:  %s\n\n", dirName1);
			fprintf(gFptr, "      FILE <%s>\n", f1->fileName);
			flag = 1;
		}
		fprintf(gFptr, "\n      MSG:21 - FILE ATTRIBUTE CHANGE\n");
		fprintf(gFptr, "      file1:%08x\n      file2:%08x\n", f1->attrib, f2->attrib);
	}

	if(strcmp(f1->created, f2->created) != 0)
	{
		if(flag == 0)
		{
			fprintf(gFptr, "\n********************************\n  <8> FILE PROPERTY MISMATCH:  %s\n\n", dirName1);
			fprintf(gFptr, "      FILE <%s>\n", f1->fileName);
			flag = 1;
		}
		fprintf(gFptr, "\n      <9> - FILE CREATE DATE DIFFERENT\n");
		fprintf(gFptr, "      file1:%s %s\n      file2:%s %s\n", 
			f1->created, f1->createdTime, f2->created, f2->createdTime);
	}
	//TIME
	else	// if dates are the same, check the times
	if(strcmp(f1->createdTime, f2->createdTime) != 0)
	{
		if(flag == 0)
		{
			fprintf(gFptr, "\n********************************\n  <A> FILE PROPERTY MISMATCH:  %s\n\n", dirName1);
			fprintf(gFptr, "      FILE <%s>\n", f1->fileName);
			flag = 1;
		}
		fprintf(gFptr, "\n      <B> FILE CREATE TIME DIFFERENT\n");
		fprintf(gFptr, "      file1:%s %s\n      file2:%s %s\n", 
			f1->created, f1->createdTime, f2->created, f2->createdTime);
	}
	//TIME

	if(strcmp(f1->modified, f2->modified) != 0)
	{
		if(flag == 0)
		{
			fprintf(gFptr, "\n********************************\n  <C> FILE PROPERTY MISMATCH:  %s\n\n", dirName1);
			fprintf(gFptr, "      FILE <%s>\n", f1->fileName);
			flag = 1;
		}
		fprintf(gFptr, "\n      <D> - FILE MODIFY DATE DIFFERENT\n");
		fprintf(gFptr, "      file1:%s %s\n      file2:%s %s\n", 
			f1->modified, f1->modifiedTime, f2->modified, f2->modifiedTime);
	}
	//TIME
	else
	if(strcmp(f1->modifiedTime, f2->modifiedTime) != 0)
	{
		if(flag == 0)
		{
			fprintf(gFptr, "\n********************************\n  <E> FILE PROPERTY MISMATCH:  %s\n\n", dirName1);
			fprintf(gFptr, "      FILE <%s>\n", f1->fileName);
			flag = 1;
		}
		fprintf(gFptr, "\n      <F> - FILE MODIFY TIME DIFFERENT\n");
		fprintf(gFptr, "      file1:%s %s\n      file2:%s %s\n", 
			f1->modified, f1->modifiedTime, f2->modified, f2->modifiedTime);
	}
//*/

/*	FILE ACCESS TIME
	// access date not good indicator here since Anit-Virus scanner touches all files
	if(strcmp(f1->accessed, f2->accessed) != 0)
	{
		if(flag == 0)
		{
			fprintf(gFptr, "\n********************************\nFILE PROPERTY MISMATCH:  %s\n\n", dirName1);
			fprintf(gFptr, "      FILE <%s>\n", f1->fileName);
			flag = 1;
		}
		fprintf(gFptr, "\n        MSG:30 - FILE ACCESS DATE DIFFERENT\n");
		fprintf(gFptr, "        file1:%s %s\n        file2:%s %s\n", 
			f1->accessed, f1->accessedTime, f2->accessed, f2->accessedTime);
	}
	//TIME
	else
	if(strcmp(f1->accessedTime, f2->accessedTime) != 0)
	{
		if(flag == 0)
		{
			fprintf(gFptr, "\n********************************\nFILE PROPERTY MISMATCH:  %s\n\n", dirName1);
			fprintf(gFptr, "      FILE <%s>\n", f1->fileName);
			flag = 1;
		}
		fprintf(gFptr, "\n        MSG:31 - FILE ACCESS TIME DIFFERENT\n");
		fprintf(gFptr, "        file1:%s %s\n        file2:%s %s\n", 
			f1->accessed, f1->accessedTime, f2->accessed, f2->accessedTime);
	}
	// TIME
//*/

	if(flag == 1) fprintf(gFptr, "\n");
	return;
} // matchFileAttributes

// check file list for differences
void matchFiles(DIRNODE *d1, DIRNODE *d2, char *dirName1, char *dirName2)
{
	FILENODE *f1, *f2;

	f1 = d1->fileList;
	f2 = d2->fileList;
	while(f1 != NULL)
	{
		// if two files are a close match, this will be more efficient
		f2 = findFileInList(f1, f2);

		// if not found, need to search whole list
		if(f2 == NULL) f2 = findFileInList(f1, d2->fileList);

		if(f2 != NULL)	// names match
		{
			// check attributes here
			matchFileAttributes(f1, f2, dirName1, dirName2);
		}
		else
		{
			fprintf(gFptr, "\n********************************\n  <3> - EXTRA FILE IN DIR1 --> %s\n\n", dirName1);
			fprintf(gFptr, "  FILE <%s> SIZE:%lld bytes\n", f1->fileName, f1->fileSizeQ.QuadPart);
			fprintf(gFptr, "  CREATED:%s %s  MODIFIED:%s %s\n", f1->created, f1->createdTime, f1->modified, f1->modifiedTime);
			fprintf(gFptr, "\n  NOT FOUND in Dir2 %s\n", d2->dirName);
		}
		f1 = f1->next;
	}

	f2 = d2->fileList;
	while(f2 != NULL)
	{
		if(f2->checked == 0)	// not found
		{
			fprintf(gFptr, "\n********************************\n  <4> - EXTRA FILE IN DIR2 --> %s\n\n", dirName2);
			fprintf(gFptr, "  FILE <%s> SIZE:%lld bytes\n", f2->fileName, f2->fileSizeQ.QuadPart);
			fprintf(gFptr, "  CREATED:%s %s  MODIFIED:%s %s\n", f2->created, f2->createdTime, f2->modified, f2->modifiedTime);
			fprintf(gFptr, "\n  NOT FOUND in Dir1 %s\n", d1->dirName);
		}
		f2 = f2->next;
	}
	return;
} // matchFiles

// compare two log files
void matchDirs(DIRNODE *pLog1, DIRNODE *pLog2)
{
	static int numMatch = 0, level = 0, maxDepth = 0;
	static char curDir1[MAX_PATH] = { 0 }, curDir2[MAX_PATH] = { 0 };
	DIRNODE *pDir1, *pDir2;
	size_t curLen1, curLen2, prvLen1, prvLen2;

	prvLen1 = strlen(curDir1);
	prvLen2 = strlen(curDir2);

	// search log2 directories for matching log1 directories
	pDir1 = pLog1;	// iterate down the list
	while(pDir1 != NULL)
	{
		curDir1[prvLen1] = 0;	// reset directory entered to iterate through dirs at same level
		curDir2[prvLen2] = 0;

		if(pDir1->dirName != NULL)
		{
			if(strlen(curDir1) == 1)	// skip extra slash
				strcpy(curDir1, pDir1->dirName);
			else
				strcat(curDir1, pDir1->dirName);
		}
		curLen1 = strlen(curDir1);

		pDir2 = findDir(pDir1, pLog2);

		if(pDir2 != NULL)	// directories match
		{
			if(pDir2->dirName != NULL)
			{
				if(strlen(curDir2) == 1)
					strcpy(curDir2, pDir2->dirName);
				else
					strcat(curDir2, pDir2->dirName);
			}
			curLen2 = strlen(curDir2);

			numMatch++;
			level++;
			if(level > maxDepth) maxDepth = level;
			matchDirs(pDir1->subDirList, pDir2->subDirList);
			level--;
			curDir1[curLen1] = 0; // reset dir to current level
			curDir2[curLen2] = 0;
			matchFiles(pDir1, pDir2, curDir1, curDir2);
		}
		else
		{
			fprintf(gFptr, "\n********************************\n  <1> - Extra Directory \"%s\" in Dir1: %s\n\n", pDir1->dirName, curDir1);
			fprintf(gFptr, "  NOT FOUND in File2 - Missing Substructure Below:\n");
			printDirTree(gFptr, pDir1, true);
		}
		pDir1 = pDir1->next;
	} //  while

	// find any directories not matched in search of log2 by log1
	pDir2 = pLog2;
	while(pDir2 != NULL)
	{
		if(pDir2->dirName != NULL)
		{
			if(strlen(curDir2) == 1)
				strcpy(curDir2, pDir2->dirName);
			else
				strcat(curDir2, pDir2->dirName);
		}

		if(pDir2->checked == 0)	// wasn't found by other search
		{
			fprintf(gFptr, "\n********************************\n  <2> - Extra Directory \"%s\" in Dir2: %s\n\n", pDir2->dirName, curDir2);
			fprintf(gFptr, "  NOT FOUND in File1 - Missing Substructure Below:\n");
			printDirTree(gFptr, pDir2, true);
		}
		curDir2[prvLen2] = 0;
		pDir2 = pDir2->next;
	}

	if(level == 0) fprintf(gFptr, "\n\nTotal Directories Matched = %d at max depth = %d.\n\n", numMatch, maxDepth);
	return;
} // matchDirs

// compare two log files and output the difference
DIRNODE *readFileLog(FILE *fptr1)
{
	DIRNODE *pCurDirNode[1024]; // max, 1024 levels of directories
	FILENODE tmpFileNode, *pFileNode, *pCurFileNode;
	char line[256], tmpName[256];
	char newDir[] =			"Searching files in";
	char upDir[] =			"Continuing search in";
	char scanComplete[] =	" **********   File Scan complete.";
	char newSummary[] =		"Summary:";
	int level, i;

	// NULL out pointers at all levels
	for(level = 0; level < 1024; level++) pCurDirNode[level] = NULL;

	// skip the first lines, then establish the root directory
	while(1)
	{
		char *tmp;

		tmp = fgets(line, 256, fptr1);	// blank
		if(tmp == NULL)
		{
			fprintf(stderr, "\n\nError, could not find \"Searching files in\" start of log file.\n\n");
			return(NULL);
		}
		if(strncmp(line, "Searching files in", 18) == 0) break;
	}

	// set up tmp space for filename
	tmpFileNode.fileName = (char *) malloc(256);
	if(tmpFileNode.fileName == NULL)
	{
		fprintf(stderr, "\n\nError, could not allocate memory for temp file name.\n\n");
		return(NULL);
	}
	tmpFileNode.next = NULL;

	// should be root directory
	strcpy(tmpName, &line[21]);

	level = 0; // inc level to zero for initial directory
	pCurDirNode[level] = newDirNode(tmpName); // only one node at level zero
	if(pCurDirNode[level] == NULL) return(NULL);

	// read log1 file into internal memory structure
	while(!feof(fptr1))
	{
		// read the line
		fgets(line, 256, fptr1);

		// carriage return, blank line in a text file returns 10
		if(line[0] == 10) continue; 

		// error on the line
		if(strncmp(line, "Error!", 6) == 0)	continue;

		// it's a directory change down
		if(strncmp(line, newDir, 18) == 0)	// "Searching files in"
		{
			i = (int) strlen(line);
			while(i > 21)
			{
				if(line[i] == '\\') break;
				i--;
			}
			strcpy(tmpName, &line[i]);	// name at this level is "\dir name"

			// create the first sub dir node for next level
			if(pCurDirNode[level]->subDirList == NULL)
			{
				pCurDirNode[level]->subDirList = newDirNode(tmpName);// all elements set
				if(pCurDirNode[level]->subDirList == NULL) return(NULL);
				level++;
				pCurDirNode[level] = pCurDirNode[level-1]->subDirList; // set pointer at this level to current node
			}
			else	// create a new node at this level
			{
				level++;
				pCurDirNode[level]->next = newDirNode(tmpName);
				if(pCurDirNode[level]->next == NULL) return(NULL);
				pCurDirNode[level] = pCurDirNode[level]->next;
			}
			continue;
		}

		// out of files in sub directory, change up
		if(strncmp(line, upDir, 20) == 0) // "Continuing search in"
		{
			level--;
			if(level < 0)
			{
				fprintf(stderr, "\n\nError! Up more levels than down.\n\n");
				return(NULL);
			}
			continue;
		}

		// scan complete
		if(strncmp(line, scanComplete, 33) == 0) // " **********   File Scan complete."
		{
			fgets(line, 256, fptr1);	// blank
			fgets(line, 256, fptr1);	// File Search Complete
			fgets(line, 256, fptr1);	// blank
			fgets(line, 256, fptr1);	// Summary

			// total number of directories in decimal
			fgets(line, 256, fptr1);
			gTotalNumSubDirs1 = atoi(&line[35]);	// Total Number of Sub Directories

			// get total number of files
			fgets(line, 256, fptr1);
			gTotalNumFiles1 = atoi(&line[25]);		// Total Number of Files

			fgets(line, 256, fptr1);
			gTotalFileSize1 = _atoi64(&line[23]);	// Total Size of Files
			fgets(line, 256, fptr1); // get last blank line

			continue; // should be end of file
		} // if scan complete

		// must be a file name
		strcpy(tmpFileNode.fileName, line);	// save file name

		fgets(line, 256, fptr1); // always followed by file data
		if(line[0] == '\t')	// it's file data
		{
			strncpy(tmpFileNode.created, &line[5], 10);
			tmpFileNode.created[10] = 0;
			strncpy(tmpFileNode.accessed, &line[20], 10);
			tmpFileNode.accessed[10] = 0;
			strncpy(tmpFileNode.modified, &line[35], 10);
			tmpFileNode.modified[10] = 0;
			tmpFileNode.fileSizeQ.QuadPart = hex2QWORD(&line[51], 8);
			tmpFileNode.attrib = hex2DWORD(&line[67], 8);

			// TIME
			fgets(line, 256, fptr1);	// read the time info next
			strncpy(tmpFileNode.createdTime, &line[5], 8);
			tmpFileNode.createdTime[8] = 0;
			strncpy(tmpFileNode.accessedTime, &line[20], 8);
			tmpFileNode.accessedTime[8] = 0;
			strncpy(tmpFileNode.modifiedTime, &line[35], 8);
			tmpFileNode.modifiedTime[8] = 0;
			// TIME

			// create a new file node
			pFileNode = newFileNode(&tmpFileNode);
			if(pFileNode == NULL) return(NULL);

			// insert onto list of files
			if(pCurDirNode[level]->fileList == NULL)	// first on list
			{
				pCurDirNode[level]->fileList = pFileNode;
			}
			else	// linear search to last file, list will usually be short
			{
				pCurFileNode = pCurDirNode[level]->fileList;
				while(pCurFileNode->next != NULL)
					pCurFileNode = pCurFileNode->next;
				pCurFileNode->next = pFileNode;
			}
			continue;
		} // if line begins with tab - file data
	} // while not eof

//printDirTree(stdout, pCurDirNode[0], true);
//fprintf(stdout, "\n\n");

	if(tmpFileNode.fileName != NULL) free(tmpFileNode.fileName);
	return(pCurDirNode[0]);
} // readFileLog

void addFile2SizeDB(WIN32_FIND_DATA *fileData)
{
	FILENODE_DB *ptrFN, *tmpFN, *prevFN;
	DWORD fileSizeBucket;

	ptrFN = getFileNode(fileData);
	if(ptrFN == NULL) return;

	fileSizeBucket = (ptrFN->fileSizeQ.QuadPart / 4096);
	if(fileSizeBucket > 1024*1024) fileSizeBucket = 1024*1024;	// normally >=, BUT we adaded one to our array size, so 1024*1024 is a valid element

	tmpFN = gFileNodeDB_SizeArray[fileSizeBucket];
	if(tmpFN == NULL)	// list for this bucket size is empty
	{
		gFileNodeDB_SizeArray[fileSizeBucket] = ptrFN;
		return;
	}
	
	// find the position on the list based on size
	prevFN = NULL;
	while(tmpFN != NULL)
	{
		if(ptrFN->fileSizeQ.QuadPart >= tmpFN->fileSizeQ.QuadPart)	// insert here
		{
			if(prevFN == NULL)	// first on list
			{
				gFileNodeDB_SizeArray[fileSizeBucket] = ptrFN;
			}
			else
			{
				prevFN->nextFile = ptrFN;
			}
			ptrFN->nextFile = tmpFN;
			return;
		}
		prevFN = tmpFN;
		tmpFN = tmpFN->nextFile;
	}

	// to get here, tmpFN MUST be NULL
	prevFN->nextFile = ptrFN;

	return;
} // addFile2SizeDB

// recurse through directories
void recurseFindFile()
{
	static DIRNODE *ptrTopDirNode = NULL;
	static FILENODE *fileSizeArray[0x100];

	BOOL result = TRUE;
	WIN32_FIND_DATA fileData;
	HANDLE validFile;
	TCHAR dot[4] = ".", ddot[4] = "..";
	size_t dirLength;
	SYSTEMTIME created, accessed, modified;
	FILETIME cTime, aTime, mTime;
	DWORD tmp;

	dirLength = strlen(gCurrentPath);
	fprintf(gFptr, "\nSearching files in %s\n\n", gCurrentPath);
	validFile = FindFirstFile("*.*", &fileData);

	while(result)
	{
		// if it's a directory, may need to go into it
		if( (fileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0)
		{
			// if it's NOT a '.' or '..' recurse down into it
			if(strcmp(fileData.cFileName, dot) != 0 && strcmp(fileData.cFileName, ddot) != 0)
			{
				gTotalNumSubDirs1++;
				if(strlen(gCurrentPath) != 3) // if at root, do not add '\'
				{
					strcat(gCurrentPath, "\\");
				}
				strcat(gCurrentPath, fileData.cFileName);	// add the found directory to the path
				tmp = SetCurrentDirectory(gCurrentPath);

				if(tmp != 0)
				{
					recurseFindFile();	// set directory successful

					// reset directory to what it was when we came in here
					// don't need to call SetCurDir() function because we already have a valid handle
					gCurrentPath[dirLength] = 0;
					fprintf(gFptr, "\nContinuing search in %s\n\n", gCurrentPath); // only want this message when SetCurDir was successful
									// this message tells the compare log function to go up
				}
				else
				{
					DWORD err = GetLastError();
					fprintf(stderr, "\nError! Could not set directory to \"%s\", because err=%d\n\n", gCurrentPath, err);
					fprintf(gFptr, "\nError! Could not set directory to \"%s\", because err=%d\n\n", gCurrentPath, err);
					gCurrentPath[dirLength] = 0;	// reset directory name
				}

			} // end recursed into a directory
			//else
			//	printf("DIRECTORY!!!!!\n\n");
		}
		else
		{
			// if we need to sort at the end, then add nodes to the list
			if(gSearchFileDate)
			{
			}

			if(gSearchFileSize)
			{
				addFile2SizeDB(&fileData);
			}

			countAttribs(fileData.dwFileAttributes);
			gTotalNumFiles1++;
			gTotalFileSize1 += fileData.nFileSizeLow + (QWORD) fileData.nFileSizeHigh * 0x100000000;
			memset(&created, 0, sizeof(SYSTEMTIME));
			memset(&accessed, 0, sizeof(SYSTEMTIME));
			memset(&modified, 0, sizeof(SYSTEMTIME));

			FileTimeToLocalFileTime(&fileData.ftCreationTime, &cTime);
			FileTimeToLocalFileTime(&fileData.ftLastAccessTime, &aTime);
			FileTimeToLocalFileTime(&fileData.ftLastWriteTime, &mTime);

			FileTimeToSystemTime(&cTime, &created);
			FileTimeToSystemTime(&aTime, &accessed);
			FileTimeToSystemTime(&mTime, &modified);

			if(created.wYear > 9999) created.wYear = 9999;		// some instances where this value exceeded allowed size
			if(accessed.wYear > 9999) accessed.wYear = 9999;	// some instances where this value exceeded allowed size
			if(modified.wYear > 9999) modified.wYear = 9999;	// some instances where this value exceeded allowed size

			//FileTimeToSystemTime(&fileData.ftCreationTime, &created);
			//FileTimeToSystemTime(&fileData.ftLastAccessTime, &accessed);
			//FileTimeToSystemTime(&fileData.ftLastWriteTime, &modified);

			// Created, Accessed, Modified --> DATE
			fprintf(gFptr, "%s\n", fileData.cFileName);
			fprintf(gFptr, "\tCrt:%02d/%02d/%04d Acc:%02d/%02d/%04d Mod:%02d/%02d/%04d Size:%08x Attrib:%08x\n", 
				created.wMonth, created.wDay, created.wYear,
				accessed.wMonth, accessed.wDay, accessed.wYear,
				modified.wMonth, modified.wDay, modified.wYear,
				fileData.nFileSizeLow,
				fileData.dwFileAttributes);

			//TIME
			// Created, Accessed, Modified --> TIME
			fprintf(gFptr, "\tCrt:%02d:%02d:%02d   Acc:%02d:%02d:%02d   Mod:%02d:%02d:%02d\n\n", 
				created.wHour, created.wMinute, created.wSecond,
				accessed.wHour, accessed.wMinute, accessed.wSecond,
				modified.wHour, modified.wMinute, modified.wSecond
				);
			// TIME
		}
		// get the next file
		result = FindNextFile(validFile, &fileData);
	} // while result
	FindClose(validFile);	// this must be used to close this handle, otherwise resources run out
							// got 1816 "ERROR_NOT_ENOUGH_QUOTA" from SetCurrentDirectory
	return;
} // recurseFindFile

void searchFiles(SYSTEMTIME *sysTime)
{
	char logFileName[MAX_PATH], tmpFileName[MAX_PATH];
	DWORD tmp;

	// init the attrib cnt
	for(tmp = 0; tmp < 32; tmp++)
		attribCnt[tmp] = 0;

	sprintf(logFileName, "FileList_%04d_%02d_%02d_%02d_%02d_%02d.log", 
		sysTime->wYear, sysTime->wMonth, sysTime->wDay, 
		sysTime->wHour, sysTime->wMinute, sysTime->wSecond);

	// open the file
	gFptr = fopen(logFileName, "w+");
	if(gFptr == NULL)
	{
		fprintf(stderr, "\n\nError! Could not open log file %s for writing.\n\n", logFileName);
		return;
	}

	fprintf(gFptr, "DIRECTORY\n");
	fprintf(gFptr, "File Search started on %02d/%02d/%04d at %02d:%02d:%02d\n\n", 
		sysTime->wMonth, sysTime->wDay, sysTime->wYear,
		sysTime->wHour, sysTime->wMinute, sysTime->wSecond);

	fprintf(stderr, "\nFile Search started on %02d/%02d/%04d at %02d:%02d:%02d\n\n", 
		sysTime->wMonth, sysTime->wDay, sysTime->wYear,
		sysTime->wHour, sysTime->wMinute, sysTime->wSecond);

	// Start in Directory from which executed for now
	tmp = GetCurrentDirectory(MAX_PATH + 1, gCurrentPath);
	if(tmp == 0)
	{
		fprintf(stderr, "\n\nError! Could not get current working directory.\n\n");
		return;
	}

	// find all the files and create a log file
	recurseFindFile();

	fprintf(gFptr, "\n **********   File Scan complete.\n\n");
	fprintf(stderr, "\n **********   File Scan complete.\n\n");

	printAttribs(gFptr);

	GetSystemTime(sysTime);
	fprintf(gFptr, "File Search complete on %d/%d/%d at %d:%d:%d\n\n", 
		sysTime->wMonth, sysTime->wDay, sysTime->wYear,
		sysTime->wHour, sysTime->wMinute, sysTime->wSecond);

	fprintf(gFptr, "Summary:\n  Total Number of Sub-Directories: %u\n", gTotalNumSubDirs1);
	fprintf(gFptr, "  Total Number of Files: %u\n", gTotalNumFiles1);
	fprintf(gFptr, "  Total Size of Files: %I64u\n", gTotalFileSize1);

	fprintf(stderr, "File Search complete on %d/%d/%d at %d:%d:%d\n\n", 
		sysTime->wMonth, sysTime->wDay, sysTime->wYear,
		sysTime->wHour, sysTime->wMinute, sysTime->wSecond);

	fprintf(stderr, "Summary:\n  Total Number of Sub-Directories: %u\n", gTotalNumSubDirs1);
	fprintf(stderr, "  Total Number of Files: %u\n", gTotalNumFiles1);
	fprintf(stderr, "  Total Size of Files: %I64u\n", gTotalFileSize1);

	fclose(gFptr);
	return;
} // searchFiles



/* //
DWORD mask = FILE_ATTRIBUTE_OFFLINE;
DWORD orig;
BOOL rrr;
HANDLE hDevice;
 
hDevice = CreateFile(fileData.cFileName, FILE_ALL_ACCESS, FILE_SHARE_READ, NULL, OPEN_EXISTING, 
					 FILE_ATTRIBUTE_NORMAL | FILE_FLAG_BACKUP_SEMANTICS, NULL);

orig = fileData.dwFileAttributes;
printf("File=%s, set to %08x\n", fileData.cFileName, fileData.dwFileAttributes | mask);

if(mask == FILE_ATTRIBUTE_SPARSE_FILE)
{
	DWORD numbytes;

	rrr = DeviceIoControl(hDevice, FSCTL_SET_SPARSE, NULL, 0, NULL, 0, &numbytes, NULL);
}
else
{
	rrr = SetFileAttributes(fileData.cFileName, fileData.dwFileAttributes | mask);
}
printf("Rrr = %d\n", rrr);

DWORD aaa = GetFileAttributes(fileData.cFileName);

if( (aaa & mask) != 0)
	printf("(%08x) success!\n", aaa);
else
	printf("(%08x) failure :-(\n", aaa);

//SetFileAttributes(fileData.cFileName, orig);
CloseHandle(hDevice);
//*/

