// Main.cpp
// 
// Main code for finding all files and/or registry keys
//

#include "Main.h"
#include <io.h>

TCHAR gCurrentPath[MAX_PATH + 1], gStartPath[MAX_PATH+1];;
FILE *gFptr = NULL, *gfptrRegData1 = NULL, *gfptrRegData2 = NULL;
PROCESS_NODE *gProc1 = NULL, *gProc2 = NULL;

BYTE *gpRegData1 = NULL, *gpRegData2 = NULL;
long gFileSize1 = 0, gFileSize2 = 0;

unsigned int gTotalNumFiles1 = 0, gTotalNumSubDirs1 = 0;
unsigned __int64 gTotalFileSize1 = 0;

unsigned int gTotalNumFiles2 = 0, gTotalNumSubDirs2 = 0;
unsigned __int64 gTotalFileSize2 = 0;

bool gCompareFiles = false;
bool gSearchFiles = false;
bool gSearchFileDate = false;
bool gSearchFileSize = false;

bool gSearchRegistry = false;
bool gSaveRegistryData = false;
bool gSearchServices = false;
bool gSearchProcesses = false;

DWORD gMaxFileNodeDB_Count = 512 * 1024;
DWORD gFileNodeDB_Count = 0;
FILENODE_DB *gptrTopFileNodeDB = NULL;
FILENODE_DB *gFileNodeDB_SizeArray[1024 * 1024 + 1];	// ecah bucked is 4KB file size, [1MB+1] for sizes >= 4GB

void initFileNodeMemory()	// had 330,000 files on system, let's allow for 512 * 1024
{
	printf("Initializing file node memory structures.\n\n");

	gptrTopFileNodeDB = (FILENODE_DB *) VirtualAlloc(NULL, gMaxFileNodeDB_Count * sizeof(FILENODE_DB), MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE);
	if(gptrTopFileNodeDB == NULL)
	{
		fprintf(stderr, "Memory Error! Could not allcoate %u bytes for %u file node (%u) objects\n", 
			sizeof(FILENODE_DB) * gMaxFileNodeDB_Count, gMaxFileNodeDB_Count, sizeof(FILENODE_DB));
		exit(0);
	}

	for(int i = 0; i < (1024*1024+1); i++)	// this is an optimization to sort the file list by size
		gFileNodeDB_SizeArray[i] = NULL;

	printf("Done initializing file node memory structures.\n\n");
	return;
} // initFileNodeMemory

// this function prints out the correct syntax for executing the program
void usage()
{
	printf("\n\n---------  USAGE for FootPrint %f ---------\n\n", VERSION);
	printf("FP%2d_%-4.2f.exe [option1 option2 ... option n]\n\n", BITS, VERSION);
	printf("-f (log all files in this directory and all sub-directories)\n");
	printf("  -fd (ALSO save files by newest date)\n");
	printf("  -fs (ALSO save files by largest size)\n\n");

	printf("-r (log registry entries)\n");
	printf("  -rd (log registry entries and their data)\n\n");

	printf("-s (log running services)\n\n");
	printf("-p (log running processes)\n\n");

	printf("FP%2d_%-4.2f.exe -cmp logFile1 logFile2  (compare two log files)\n\n", BITS, VERSION);
	printf("Can't use \"-cmp\" with any other options\n\n");

	return;
} // usage

// this function sets global flags
void parseCommandLine(int argc, char *argv[], char *logFile1, char *logFile2)
{
	int x;

	// arrays passed in, set first character to zero
	*logFile1 = 0;
	*logFile2 = 0;

	gCompareFiles = false;
	gSearchFiles = false;
	gSearchRegistry = false;
	gSearchServices = false;
	gSearchProcesses = false;

	x = 1;
	while(x < argc)
	{
		if(strcmp(argv[x], "-f") == 0)
		{
			if(gCompareFiles == true)
			{
				fprintf(stderr, "\n\nError - cannot search system and compare files at the same time.\n\n");
				exit(-1);
			}
			gSearchFiles = true;
		}
		else if(strcmp(argv[x], "-fd") == 0)
		{
			if(gCompareFiles == true)
			{
				fprintf(stderr, "\n\nError - cannot search system and compare files at the same time.\n\n");
				exit(-1);
			}
			gSearchFiles = true;
			gSearchFileDate = true;
		}
		else if(strcmp(argv[x], "-fs") == 0)
		{
			if(gCompareFiles == true)
			{
				fprintf(stderr, "\n\nError - cannot search system and compare files at the same time.\n\n");
				exit(-1);
			}
			gSearchFiles = true;
			gSearchFileSize = true;
		}

		else if(strcmp(argv[x], "-r") == 0)
		{
			if(gCompareFiles == true)
			{
				fprintf(stderr, "\n\nError - cannot search system and compare files at the same time.\n\n");
				exit(-1);
			}
			gSearchRegistry = true;
		}
		else if(strcmp(argv[x], "-rd") == 0)
		{
			if(gCompareFiles == true)
			{
				fprintf(stderr, "\n\nError - cannot search system and compare files at the same time.\n\n");
				exit(-1);
			}
			gSearchRegistry = true;
			gSaveRegistryData = true;
		}
		else if(strcmp(argv[x], "-s") == 0)
		{
			if(gCompareFiles == true)
			{
				fprintf(stderr, "\n\nError - cannot search system and compare files at the same time.\n\n");
				exit(-1);
			}
			gSearchServices = true;
		}
		else if(strcmp(argv[x], "-p") == 0)
		{
			if(gCompareFiles == true)
			{
				fprintf(stderr, "\n\nError - cannot search system and compare files at the same time.\n\n");
				exit(-1);
			}
			gSearchProcesses = true;
		}
		else if(strcmp(argv[x], "-cmp") == 0)
		{
			if(gSearchFiles == true || gSearchRegistry == true || gSearchServices == true)
			{
				fprintf(stderr, "\n\nError - Can't compare log files and search the system in same command.\n\n");
				exit(-1);
			}
			gCompareFiles = true;
		}
		else	// if it's not one of the other options, then it's assumed to be a log file to compare
		{
			if(*logFile1 != 0 && *logFile2 != 0)
			{
				fprintf(stderr, "\n\nError - Can't specify more than two files to compare.\n\n");
				exit(-1);
			}

			if(*logFile1 == 0)
			{
				strcpy(logFile1, argv[x]);
			}
			else if(*logFile2 == 0)
			{
				strcpy(logFile2, argv[x]);
			}
		}
		x++;
	}

	return;
} // parseCommandLine

// match file logs, determine when a directory or registry log file
void matchLogFiles(char *logFile1, char *logFile2)
{
	FILE *fptr1, *fptr2, *fptrRD;
	DIRNODE *pLog1, *pLog2;
	REG_KEY_NODE *pRegKey1, *pRegKey2;
	REG_DATA_NODE *pRegData1, *pRegData2;
	SERVICE_NODE *pSvc1, *pSvc2;

	char line[1024];
	enum _FILETYPE fileFlag1, fileFlag2;
	DWORD numSrvc1, numSrvc2, numberOfDiscrepancies;

	char cmpLogFile[MAX_PATH], regDataFileName1[MAX_PATH], regDataFileName2[MAX_PATH];


	// open the log files to compare
	fptr1 = fopen(logFile1, "r");
	if(fptr1 == NULL)
	{
		fprintf(stderr, "\n\nError! Could not open log file %s for reading.\n\n", logFile1);
		return;
	}

	fptr2 = fopen(logFile2, "r");
	if(fptr1 == NULL)
	{
		fprintf(stderr, "\n\nError! Could not open log file %s for reading.\n\n", logFile2);
		fclose(fptr1);
		return;
	}

	fileFlag1 = Undetermined;
	fileFlag2 = Undetermined;

	// check file one
	fgets(line, 1024, fptr1);
	if(strncmp(line, "REGISTRY+DATA", 13) == 0)
		fileFlag1 = RegistryData;
	else if(strncmp(line, "REGISTRY", 8) == 0)
		fileFlag1 = Registry;			// registry file
	else if(strncmp(line, "DIRECTORY", 9) == 0)
		fileFlag1 = Directory;
	else if(strncmp(line, "SERVICES", 8) == 0)
		fileFlag1 = Service;
	else if(strncmp(line, "PROCESSES", 8) == 0)
		fileFlag1 = Process;

	if(fileFlag1 == Undetermined)
	{
		fprintf(stderr, "\n\nError! File1 is not a known type.\n\n");
		exit(-1);
	}

	// check file two
	fgets(line, 1024, fptr2);
	if(strncmp(line, "REGISTRY+DATA", 13) == 0)
		fileFlag2 = RegistryData;
	else if(strncmp(line, "REGISTRY", 8) == 0)
		fileFlag2 = Registry;			// registry file
	else if(strncmp(line, "DIRECTORY", 9) == 0)
		fileFlag2 = Directory;
	else if(strncmp(line, "SERVICES", 8) == 0)
		fileFlag2 = Service;
	else if(strncmp(line, "PROCESSES", 8) == 0)
		fileFlag2 = Process;

	if(fileFlag2 == Undetermined)
	{
		fprintf(stderr, "\n\nError! File2 is not a known type.\n\n");
		exit(-1);
	}

	// both files must be of the same type
	if(fileFlag1 != fileFlag2)
	{
		if( !( (fileFlag1 == Registry && fileFlag2 == RegistryData) || (fileFlag1 == RegistryData && fileFlag2 == Registry) ) )
		{
			fprintf(stderr, "\n\nError! Files one and two are not the same type.\n\n");
			exit(-1);
		}
	}

	// create name for comparison file
	sprintf(cmpLogFile, "cmp_%s_2_%s_.log", logFile1, logFile2);

	gFptr = fopen(cmpLogFile, "w+");
	if(gFptr == NULL)
	{
		fprintf(stderr, "\n\nError! Could not open output comparison file %s for writing.\n\n", cmpLogFile);
		exit(-1);
	}

	// read directory files
	if(fileFlag1 == Directory)			// directory files, both flags must be equal at this point
	{
		fprintf(stdout, "Reading DIRECTORY file one:%s\n", logFile1);
		pLog1 = readFileLog(fptr1);
		if(pLog1 == NULL)
		{
			fclose(fptr1);
			fclose(fptr2);
			fclose(gFptr);
			return;
		}
//printDirTree(pLog1);

		// read the second file
		fprintf(stdout, "Reading DIRECTORY file two:%s\n", logFile2);
		pLog2 = readFileLog(fptr2);
		if(pLog2 == NULL)
		{
			fclose(fptr1);
			fclose(fptr2);
			fclose(gFptr);
			return;
		}
//printDirTree(pLog2);

	}
	else if(fileFlag1 == Registry || fileFlag1 == RegistryData)	// registry file
	{
		fprintf(stdout, "\nReading REGISTRY file one:%s\n", logFile1);
		pRegKey1 = readRegFile(fptr1);
		if(pRegKey1 == NULL)
		{
			fclose(fptr1);
			fclose(fptr2);
			fclose(gFptr);
			return;
		}
//printRegNodeTree(stdout, pRegKey1, 0);

		// read the second file
		fprintf(stdout, "\nReading REGISTRY file two:%s\n", logFile2);
		pRegKey2 = readRegFile(fptr2);
		if(pRegKey2 == NULL)
		{
			fclose(fptr1);
			fclose(fptr2);
			fclose(gFptr);
			return;
		}
//printRegNodeTree(stdout, pRegKey2, 0);

		if(fileFlag1 == RegistryData && fileFlag2 == RegistryData)	// if both are not true, just treat like regular registry file
		{
			// need to verify both files exist and allocate memory accordingly
			strcpy(regDataFileName1, logFile1);
			strcat(regDataFileName1, "_DF.dat");

			fptrRD = fopen(regDataFileName1, "rb");
			if(fptrRD == NULL)	// can't read data file 1
			{
				goto RegDataFileError;
			}

			gFileSize1 = _filelength( fileno(fptrRD) );
			gpRegData1 = (BYTE *) malloc(gFileSize1);
			if(gpRegData1 == NULL)	// can't allocate RAM for data file 1
			{
				fclose(fptrRD);
				goto RegDataFileError;
			}
			fread(gpRegData1, 1, gFileSize1, fptrRD);
			fclose(fptrRD);

			// read file 2 data
			strcpy(regDataFileName2, logFile2);
			strcat(regDataFileName2, "_DF.dat");

			fptrRD = fopen(regDataFileName2, "rb");
			if(fptrRD == NULL)
			{
				free(gpRegData1);
				goto RegDataFileError;
			}

			gFileSize2 = _filelength( fileno(fptrRD) );
			gpRegData2 = (BYTE *) malloc(gFileSize2);
			if(gpRegData2 == NULL)
			{
				free(gpRegData1);
				fclose(fptrRD);
				goto RegDataFileError;
			}
			fread(gpRegData2, 1, gFileSize2, fptrRD);
			fclose(fptrRD);

			gSaveRegistryData = true;	// made it through all the wickets
		}
	}
	else if(fileFlag1 == Service)
	{
		fprintf(stdout, "\nReading SERVICE file one:%s\n", logFile1);
		pSvc1 = readServiceFile(fptr1, numSrvc1);
		if(pSvc1 == NULL)
		{
			fclose(fptr1);
			fclose(fptr2);
			fclose(gFptr);
			return;
		}

		// read the second file
		fprintf(stdout, "\nReading SERVICE file two:%s\n", logFile2);
		pSvc2 = readServiceFile(fptr2, numSrvc2);
		if(pSvc2 == NULL)
		{
			fclose(fptr1);
			fclose(fptr2);
			fclose(gFptr);
			return;
		}
	}

	else if(fileFlag1 == Process)
	{
		fprintf(stdout, "\nReading PROCESS file one:%s\n", logFile1);
		gProc1 = readProcessFile(fptr1);
		if(gProc1 == NULL)
		{
			fclose(fptr1);
			fclose(fptr2);
			fclose(gFptr);
			return;
		}

		// read the second file
		fprintf(stdout, "\nReading PROCESS file two:%s\n", logFile2);
		gProc2 = readProcessFile(fptr2);
		if(gProc2 == NULL)
		{
			fclose(fptr1);
			fclose(fptr2);
			fclose(gFptr);
			return;
		}
	}

// this is used to skip reading registry data files
RegDataFileError:

	// compare the two files
	fprintf(stdout, "\nComparing %s to %s ... \n", logFile1, logFile2);

	if(fileFlag1 == Directory)
	{
		matchDirs(pLog1, pLog2);
	}
	else if(fileFlag1 == Registry || fileFlag1 == RegistryData)
	{
		char regPath1[2048], regPath2[2048];

		fprintf(stdout, "\nMatching Registry log files.\n\n");
		regPath1[0] = 0; regPath2[0] = 0;
		numberOfDiscrepancies = match2Registries(gFptr, pRegKey1, pRegKey2, 0, regPath1, regPath2, logFile1, logFile2);
		fprintf(gFptr, "\n\n * * * * *   Total Number of Registry Discrepancies: %u   * * * * * \n\n", numberOfDiscrepancies);
	}
	else if(fileFlag1 == Service)
	{
		fprintf(stdout, "\nMatching Service log files.\n\n");
		matchServices(gFptr, pSvc1, numSrvc1, pSvc2, numSrvc2, logFile1, logFile2);
	}

	else if(fileFlag1 == Process)
	{
		fprintf(stdout, "\nMatching Process log files.\n\n");
		matchProcesses(gFptr, gProc1, gProc2, logFile1, logFile2);
		fflush(gFptr);
	}

	fclose(gFptr);
	fclose(fptr1);
	fclose(fptr2);
	return;
} // matchLogFiles

// main function
void main(int argc, char *argv[])
{
	SYSTEMTIME sysTime;
	char logFile1[MAX_PATH], logFile2[MAX_PATH];
	int i;

	// if no arguments, do file search
	if(argc == 1)
	{
		usage();
		return;
	}

	fprintf(stderr, "\n\nStarting FootPrint Version:%5.2f\n\n", VERSION);

	GetSystemTime(&sysTime);
	GetCurrentDirectory(MAX_PATH, gStartPath);

	// log file names NULL terminated
	parseCommandLine(argc, argv, logFile1, logFile2);	// set option flags

	if(gSearchFileDate || gSearchFileSize)
	{
		initFileNodeMemory();	// sets up memory for file nodes for sorting, sets gptrTopFileNode
	}

	// -f option, find all files
	if(gSearchFiles == true)
	{
		FILENODE_DB *ptrFN;
		FILE *fptrSizeDB, *fptrDateDB;
		char fileNameSizeDB[MAX_PATH], fileNameDateDB[MAX_PATH];

		if(gSearchFileSize)	// open file now so it's open in running directory
		{
			sprintf(fileNameSizeDB, "FileSizeList_%04d_%02d_%02d_%02d_%02d_%02d.txt", 
				sysTime.wYear, sysTime.wMonth, sysTime.wDay, 
				sysTime.wHour, sysTime.wMinute, sysTime.wSecond);

			// open the file
			fptrSizeDB = fopen(fileNameSizeDB, "w+");
		}

		if(gSearchFileDate)
		{
			sprintf(fileNameDateDB, "FileSizeList_%04d_%02d_%02d_%02d_%02d_%02d.txt", 
				sysTime.wYear, sysTime.wMonth, sysTime.wDay, 
				sysTime.wHour, sysTime.wMinute, sysTime.wSecond);

			// open the file
			fptrDateDB = fopen(fileNameDateDB, "w+");
		}

		searchFiles(&sysTime);

		if(gSearchFileSize)
		{
			if(fptrSizeDB != NULL)
			{
				for(i = 1024*1024; i >= 0; i--)
				{
					ptrFN = gFileNodeDB_SizeArray[i];
					while(ptrFN != NULL)
					{
						fprintf(fptrSizeDB, "%012llu) %s\n", ptrFN->fileSizeQ, ptrFN->filePathName);
						ptrFN = ptrFN->nextFile;
					}
					if(gFileNodeDB_SizeArray[i] != NULL) fprintf(fptrSizeDB, "\n");	// space between each block of files
				}
			}
			else
			{
				fprintf(stderr, "Error! Could not open %s for writing.\n\n", fileNameSizeDB);
			}
		} // end if size DB

		if(gSearchFileDate)
		{
		} // end if date DB
	}

	// -r option, find all registry entries
	SetCurrentDirectory(gStartPath);
	if(gSearchRegistry == true)
		searchRegistry(&sysTime);

	// -s option, find all services
	SetCurrentDirectory(gStartPath);
	if(gSearchServices == true)
		enumerateLocalServices(&sysTime);

	// -p option, find all processes
	SetCurrentDirectory(gStartPath);
	if(gSearchProcesses == true)
		enumerateLocalProcesses(&sysTime);

	// -cmp option, compare two log files
	SetCurrentDirectory(gStartPath);
	if(gCompareFiles == true)
	{
		if(logFile1[0] != 0 && logFile2[0] != 0)
			matchLogFiles(logFile1, logFile2);
	}

	return;
} // main

