// Main.h
//
// Main header file for FootPrint program
//

#include <windows.h>
#include <winbase.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <time.h>
#include <string.h>
#include <psapi.h>

#define VERSION 2.40
#define BITS 32

// version 2.30 adds comparing data values
// 2.31 adds the printout of a file compare to show the entire directory structure when a directory is deleted/added

// to access 32-bit registry
//#define CROSS_ACCESS KEY_WOW64_32KEY

// to access 64-bit registry
#define CROSS_ACCESS KEY_WOW64_64KEY

//#define CROSS_ACCESS 0

#define MAX_REG_LEVEL 528

typedef unsigned __int64 QWORD;

// STRUCTURES
typedef struct _FILENODE_DB	// used for keeping info for sorting
{
	char filePathName[MAX_PATH];	// inlcudes the directory path too

	WORD fileNameIndex;		// index to starting point in path with just the file's name

	// date/time information
	WORD year;
	WORD month;
	WORD day;
	WORD hour;
	WORD minute;
	WORD second;
	WORD milli;

	ULARGE_INTEGER fileSizeQ;	// handle file sizes larger than 4 GB

	struct _FILENODE_DB *nextFile;

} FILENODE_DB;

typedef struct _fileNode
{
	char *fileName;
	char created[12];
	char accessed[12];
	char modified[12];
	// TIME
	char createdTime[10];
	char accessedTime[10];
	char modifiedTime[10];

	ULARGE_INTEGER fileSizeQ;	// handle file sizes larger than 4 GB

	DWORD attrib;
	int checked;
	struct _fileNode *next;
} FILENODE;

typedef struct _dirNode
{
	char *dirName;
	int checked;
	FILENODE		*fileList;
	struct _dirNode *subDirList;
	struct _dirNode *next;
} DIRNODE;

typedef struct _REG_DATA_NODE
{
	char *regDataName;		// value name
	DWORD dataType;
	DWORD dataLength;		// if non-zero, then reg VALUE
	BYTE *pdata;

	struct _REG_DATA_NODE *nextData;

} REG_DATA_NODE;


typedef struct _REG_KEY_NODE
{
	char *regKeyName;	// key name or value name
	char dateTime[24];

	bool alreadyMatched;

	struct _REG_KEY_NODE *subKey;
	struct _REG_KEY_NODE *nextKey;

	REG_DATA_NODE *regDataNode;
} REG_KEY_NODE;

enum _FILETYPE {Undetermined, Directory, Registry, RegistryData, Service, Process};

typedef struct _SERVICE_NODE	// implemented as array
{
	char svcName[256];
	char disName[256];

	DWORD svcState;
	DWORD svcType;
	DWORD svcPid;
	DWORD svcFlags;

	bool alreadyMatched;
} SERVICE_NODE;

typedef struct _NAME_LIST
{
	char name[MAX_PATH];
	DWORD address;
	bool matched;
	struct _NAME_LIST *nextName;
} NAME_LIST;

typedef struct _PROCESS_NODE
{
	DWORD pid;
	DWORD numMods;
	NAME_LIST *moduleList;	// first item is process itself

	struct _PROCESS_NODE *nextProcNode;
} PROCESS_NODE;


// PROTOTYPES

// regFind.cpp
void recurseRegKeys(HKEY hKey);
void searchRegistry(SYSTEMTIME *sysTime);
REG_KEY_NODE *readRegFile(FILE *fptr);
DWORD match2Registries(FILE *fptr, REG_KEY_NODE *ptopReg1, REG_KEY_NODE *ptopReg2, int level, char *path1, char *path2, char *file1, char *file2);

void printRegNodeTree(FILE *fptr, REG_KEY_NODE *pReg, int level);
void printRegNodeSubTree(FILE *fptr, REG_KEY_NODE *pReg, int level);

void QueryKey(HKEY hKey);

// FileFinder.cpp
void recurseFindFile();
void searchFiles(SYSTEMTIME *sysTime);
void matchLogFiles(char *logFile1, char *logFile2);
DIRNODE *readFileLog(FILE *fptr1);
void matchDirs(DIRNODE *pLog1, DIRNODE *pLog2);
DWORD hex2QWORD(char *data, int len);

// Services.cpp
void enumerateLocalServices(SYSTEMTIME *sysTime);
SERVICE_NODE *readServiceFile(FILE *fptr, DWORD &length);
void matchServices(FILE *fptr, SERVICE_NODE *pSvc1, DWORD numOfSvc1, SERVICE_NODE *pSvc2, DWORD numOfSvc2, char *logFile1, char *logFile2);

// Process.cpp
void enumerateLocalProcesses(SYSTEMTIME *sysTime);
PROCESS_NODE *readProcessFile(FILE *fptr);
void matchProcesses(FILE *fptr, PROCESS_NODE *proc1, PROCESS_NODE *proc2, char *logFile1, char *logFile2);

// GLOBAL VARIABLES
extern FILE *gFptr, *gfptrRegData1, *gfptrRegData2;
extern TCHAR gCurrentPath[MAX_PATH + 1], gStartPath[MAX_PATH+1];
extern PROCESS_NODE *gProc1, *gProc2;
extern bool gSaveRegistryData;

extern bool gSearchFileDate;
extern bool gSearchFileSize;

extern unsigned int gTotalNumFiles1, gTotalNumSubDirs1;
extern unsigned __int64 gTotalFileSize1;

extern unsigned int gTotalNumFiles2, gTotalNumSubDirs2;
extern unsigned __int64 gTotalFileSize2;

extern BYTE *gpRegData1, *gpRegData2;
extern long gFileSize1, gFileSize2;

extern DWORD gMaxFileNodeDB_Count;
extern DWORD gFileNodeDB_Count;
extern FILENODE_DB *gptrTopFileNodeDB;
extern FILENODE_DB *gFileNodeDB_SizeArray[1024 * 1024 + 1];	// ecah bucked is 4KB file size
