// Process.cpp
//
// this file enumerates existing processes on the computer
//

#include "Main.h"

DWORD gCurProcId;

// create a new process node
PROCESS_NODE *newProcesNode()
{
	PROCESS_NODE *p;

	p = (PROCESS_NODE *) malloc(sizeof(PROCESS_NODE));
	if(p == NULL)
	{
		fprintf(stderr, "\n\nFatal Memory Error, could not allocate %u bytes for PROCESS_NODE.\n\n", 
			sizeof(PROCESS_NODE) );
		exit(-1);
	}

	p->pid = 0;
	p->numMods = 0;
	p->moduleList = NULL;
	p->nextProcNode = NULL;

	return(p);
} // newProcesNode

// create a name list node
NAME_LIST *newNameList()
{
	NAME_LIST *p;

	p = (NAME_LIST *) malloc(sizeof(NAME_LIST));
	if(p == NULL)
	{
		fprintf(stderr, "\n\nFatal Memory Error, could not allocate %u bytes for NAME LIST node.\n\n",
			sizeof(NAME_LIST) );
		exit(-1);
	}

	p->address = 0;
	p->name[0] = 0;
	p->matched = false;
	p->nextName = NULL;

	return(p);
} // newNameList

// search process module list for a name
PROCESS_NODE *findProcessName(char *name, PROCESS_NODE *proc)
{
	while(proc != NULL)
	{
		// looking for the process name, should be first on the list
		if(proc->moduleList != NULL && proc->pid != 0)
		{
			if(strcmp(name, proc->moduleList->name) == 0) return(proc);
		}
		proc = proc->nextProcNode;
	}
	return(NULL); //not found
} // findProcessName

// compares two module lists
void matchModuleList(FILE *fptr, NAME_LIST *p1, NAME_LIST *p2, char *logFile1, char *logFile2, DWORD &numberOfDiscrepancies)
{
	NAME_LIST *tmp1, *tmp2;
	bool processPrintFlag;

	processPrintFlag = true;	// when true, print the process name
	tmp1 = p1;
	while(tmp1 != NULL)	// already checked first names on each of these, they match, but want to check addresses
	{
		tmp2 = p2;	// start at top of list #2
		while(tmp2 != NULL)
		{
			if(tmp2->matched == true)	// already seen this one
			{
				tmp2 = tmp2->nextName;
				continue;
			}

			if(strcmp(tmp1->name, tmp2->name) == 0)	// module names match
			{
				tmp1->matched = true;
				tmp2->matched = true;

				// check address of matching modules
				if(tmp1->address != tmp2->address)	// p1 is a NMAE LIST, but first name is also process name
				{
					if(processPrintFlag)
					{
						processPrintFlag = false;
						fprintf(fptr, "    Process: \"%s\"\n\n", p1->name);
					}
					fprintf(fptr, "    %3d) Address of Module: \"%s\"\n\n         in file1: %s\n\n"
						"         -- File1 Address (%08x) does not match File2 Address (%08x)\n\n         in Module: \"%s\"\n\n         in file2: %s\n\n", ++numberOfDiscrepancies,
						tmp1->name, logFile1, tmp1->address, tmp2->address, tmp2->name, logFile2);
				}
				break;
			}
			tmp2 = tmp2->nextName;
		} // end while looping through list2

		if(tmp2 == NULL)	// never found a module match
		{
			if(processPrintFlag)
			{
				processPrintFlag = false;
				fprintf(fptr, "    Process: \"%s\"\n\n", p1->name);
			}
			fprintf(fptr, "    %3d) Module Unloaded --> Module: \"%s\" \n\n         -- not found in file2: %s\n\n", ++numberOfDiscrepancies, tmp1->name, logFile2);
		}

		tmp1 = tmp1->nextName;
	} // while looping through list1

	// iterate through the module list of the process from file 2
	tmp2 = p2;
	while(tmp2 != NULL)
	{
		if(tmp2->matched == true)	// if matched, no need to check
		{
			tmp2 = tmp2->nextName;
			continue;
		}

		if(processPrintFlag)
		{
			processPrintFlag = false;
			fprintf(fptr, "    Process: \"%s\"\n\n", p1->name);
		}
		fprintf(fptr, "    %3d) New Module --> Module: \"%s\"\n\n         -- not found in file1: %s\n\n", ++numberOfDiscrepancies, tmp2->name, logFile2);

		tmp2 = tmp2->nextName;
	} // end while looping through list 2

	if(!processPrintFlag) fprintf(fptr, "\n");

	return;
} // matchModuleList

// match two process files
void matchProcesses(FILE *fptr, PROCESS_NODE *proc1, PROCESS_NODE *proc2, char *logFile1, char *logFile2)
{
	PROCESS_NODE *tmpP;
	NAME_LIST *tmpMod2, *pMod1;
	DWORD procUnknownCount1, procUnknownCount2, numberOfDiscrepancies;

	procUnknownCount1 = 0;
	procUnknownCount2 = 0;
	numberOfDiscrepancies = 0;

	// will search and match the first list, and will remove any matches from the 2nd list
	while(proc1 != NULL)
	{
		pMod1 = proc1->moduleList;	// first one is key, find same process in proc2
		if(pMod1 != NULL) // this process was readable in userland space
		{
			tmpP = findProcessName(pMod1->name, proc2);	// searches file 2 process list to find this file 1 process
			if(tmpP == NULL)
			{
				fprintf(fptr, "%3d) Process Done --> process: \"%s\"\n\n     -- not found in file2: %s\n\n\n", ++numberOfDiscrepancies, pMod1->name, logFile2);
				proc1 = proc1->nextProcNode;
				continue;
			}

			// tmpP is in file2, found matching process, now compare module list
			tmpMod2 = tmpP->moduleList;	// pMod is module list for proc1

			// already checked pMod and tmpN for NULL
			matchModuleList(fptr, pMod1, tmpMod2, logFile1, logFile2, numberOfDiscrepancies);

			// mark tmpP as already matched
			tmpP->pid = 0;
		}
		else	// system process we could not see
		{
			procUnknownCount1++;
		}

		// check next process on list
		proc1 = proc1->nextProcNode;
	} // end while looping through proc1

	// check the 2nd process list for anything not found in the first process list
	while(proc2 != NULL)
	{
		if(proc2->moduleList == NULL || proc2->pid == 0)
		{
			if(proc2->moduleList == NULL) procUnknownCount2++;

			proc2 = proc2->nextProcNode;
			continue;
		}
		fprintf(fptr, "%3d) New Process --> process: \"%s\"\n\n     -- not found in file1: %s\n\n", ++numberOfDiscrepancies, proc2->moduleList->name, logFile1);
		fflush(fptr);
		proc2 = proc2->nextProcNode;
	}

	if(procUnknownCount1 != procUnknownCount2)
	{
		fprintf(fptr, "%3d) Number of System Processes (%u) in file1: %s\n\n     -- not equal to\n\n     -- Number of System Process (%u) in file2: %s\n\n", 
			++numberOfDiscrepancies, procUnknownCount1, logFile1, procUnknownCount2, logFile2);
	}
	fprintf(fptr, "\n\n * * * * *   Total Number of Process Discrepancies: %u   * * * * * \n\n", numberOfDiscrepancies);
	return;
} // matchProcesses

/////////////////////
// READING LOG FILES
/////////////////////
// skip over the lines that have info on this program
void skipSelf(FILE *fptr)
{
	char buf[512];

	while(1)
	{
		fgets(buf, 512, fptr);
		if(buf[0] == '#') return;
	}
	return;
} // skipSelf

// reads the module list of a process
NAME_LIST *readModuleList(FILE *fptr, DWORD &count)
{
	NAME_LIST *pTopName, *pTmpName, *pCurName;
	char buf[512];
	DWORD x;

	count = 0;
	pTopName = NULL;
	while(1)
	{
		fgets(buf, 512, fptr);
		if(feof(fptr)) return(pTopName);
		if(buf[0] == 0x0a || buf[0] == 0x0d) return(pTopName);

		pTmpName = newNameList();	// checks for NULL at creation
		count++;

		pTmpName->address = strtol(buf, NULL, 16); // get the address
		strncpy(pTmpName->name, &buf[11], MAX_PATH);
		
		x = (DWORD) strlen(pTmpName->name)-1;
		while(x > 0)
		{
			if(pTmpName->name[x] == 0x0a || pTmpName->name[x] == 0x0d)	// NULLS the CR/LF
				pTmpName->name[x] = 0;
			else break;

			x--;
		}
		
		// add it to the list
		if(pTopName == NULL)
		{
			pTopName = pTmpName;
			pCurName = pTmpName;
		}
		else
		{
			pCurName->nextName = pTmpName;
			pCurName = pTmpName;
		}
	}
    return(pTopName);
} // readModuleList

// read the process file
PROCESS_NODE *readProcessFile(FILE *fptr)
{
	PROCESS_NODE *pTopProc, *pCurProc, *ptmp;
	char buffer[512], *pc;
	DWORD cntMods;

	pTopProc = NULL;
	pCurProc = NULL;

	while(1)
	{
		fgets(buffer, 512, fptr);
		if(feof(fptr)) break;
		if(buffer[0] == 0x0a || buffer[0] == 0x0d) continue;

		// want to skip self - don't care if features of this program are changed
		if(buffer[0] == '#')
		{
			skipSelf(fptr);
			continue;
		}

		pc = strstr(buffer, "PID");
		if(pc == NULL)
		{
			fprintf(stderr, "\n\nError, PID not found in Process file as expected.\n\n");
			continue;
		}
		ptmp = newProcesNode();
		ptmp->pid = atol(pc+4);

		// add to the list
		if(pTopProc == NULL)
		{
			pTopProc = ptmp;
			pCurProc = ptmp;
		}
		else
		{
			pCurProc->nextProcNode = ptmp;
			pCurProc = ptmp;
		}

		pc = strstr(buffer, "<Unknown>");
		if(pc != NULL) continue;	// no info available

		pc = strstr(buffer, "( ");
		ptmp->numMods = atol(pc+2);	// read the number of modules

		ptmp->moduleList = readModuleList(fptr, cntMods);	// get all the modules for this process

		if(ptmp->moduleList == NULL)
		{
			fprintf(stderr, "No modules for pid=%u\n", ptmp->pid);
		}

		else if(ptmp->numMods != cntMods)
		{
			fprintf(stderr, "\n\nNumber of modules listed (%u) for process=%s\n  not equal to number read (%u).\n\n",
				ptmp->numMods, ptmp->moduleList->name, cntMods);
			ptmp->numMods = cntMods;
		}
	} // end while looping through process file

	return(pTopProc);
} // readProcessFile

// prints process name
void printProcessNameAndModules( FILE *fptr, DWORD processID )
{
	HMODULE hMods[1024];
	HANDLE hProcess;
	DWORD cbNeeded, numMods, i;
	char szProcessName[MAX_PATH], szModName[MAX_PATH];

	// Get a handle to the process.
	hProcess = OpenProcess( PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, processID );

	// Get the process name.
	if(hProcess != NULL )
	{
		if(processID == gCurProcId)
			fprintf(fptr, "#####  START CURRENT PROCESS INFO  #####\n");

		if( EnumProcessModules( hProcess, hMods, sizeof(hMods), &cbNeeded) )
		{
			numMods = cbNeeded / sizeof(HMODULE);

			GetModuleBaseName( hProcess, hMods[0], szProcessName, sizeof(szProcessName)/sizeof(char) );
			fprintf(fptr, "PID:%04u ( %u Modules )  <%s>\n", processID, numMods, szProcessName);

			for ( i = 0; i < numMods; i++ )
			{
				// Get the full path to the module's file.
				if ( GetModuleFileNameEx( hProcess, hMods[i], szModName, sizeof(szModName)))
				{
					// Print the module name and handle value.
					fprintf(fptr, "\t%p  %s\n", hMods[i], szModName );
				}
			} // end for loop through modules

			if(processID == gCurProcId)
				fprintf(fptr, "#####  END CURRENT PROCESS INFO  #####\n");
			fprintf(fptr, "\n");
		}
	} // end if hProcess
	else
	{
		fprintf(fptr, "PID:%04u - <Unknown>\n\n", processID);
	}
	CloseHandle( hProcess );
	return;
} // printProcessNameAndModules

// enumerates all processes on the system
void enumerateLocalProcesses(SYSTEMTIME *sysTime)
{
    // Get the list of process identifiers.
    DWORD aProcesses[1024], cbNeeded, cProcesses;
    unsigned int i;

	char logFileName[MAX_PATH];

	sprintf(logFileName, "ProcessList_%04d_%02d_%02d_%02d_%02d_%02d.log", 
		sysTime->wYear, sysTime->wMonth, sysTime->wDay, 
		sysTime->wHour, sysTime->wMinute, sysTime->wSecond);

	// open the file
	gFptr = fopen(logFileName, "w+");
	if(gFptr == NULL)
	{
		fprintf(stderr, "\n\nError! Could not open log file %s for writing.\n\n", logFileName);
		return;
	}

	fprintf(stdout, "\nEnumerating Processes\n\n");
	fprintf(gFptr, "PROCESSES\n\n");

	if ( !EnumProcesses( aProcesses, sizeof(aProcesses), &cbNeeded ) ) return;

	gCurProcId = GetCurrentProcessId();

    // Calculate how many process identifiers were returned.
    cProcesses = cbNeeded / sizeof(DWORD);

    // Print the name for each process.
    for ( i = 0; i < cProcesses; i++ )
	{
        if( aProcesses[i] != 0 )
            printProcessNameAndModules( gFptr, aProcesses[i] );
	}

	fclose(gFptr);
	return;
} // enumerateLocalProcesses