// Wrappers.cpp
//
// This program will wrap any file inside a valid header for another file type
//

#include "Wrappers.h"

char *readDataFile(DWORD *size)
{
	FILE *fptrIn;
	DWORD x;
	char *data;

	// open the source file
	fptrIn = fopen(gSourceFile, "rb");
	if(fptrIn == NULL)
	{
		fprintf(stderr, "\n\nError! Could not open source file \"%s\"\n\n", gSourceFile);
		return(NULL);
	}
	*size = filelength(_fileno(fptrIn));

	data = (char *) malloc(*size);
	if(data == NULL)
	{
		fprintf(stderr, "\n\nError! Could not allocate %d bytes for source file.\n\n", *size);
		fclose(fptrIn);
		return(NULL);
	}
	x = (DWORD) fread(data, 1, *size, fptrIn); // read in the source file

	if(x != *size)
	{
		fprintf(stderr, "\n\nWarning! Tried to read %d bytes, but only read %d bytes.\n\n", 
			*size, x);
	}

	fclose(fptrIn);
	return(data);
} // readDataFile

// wraps the source file in a TXT format
// this is just a matter of changing the extension
void wrapInTXT()
{
	FILE *fptrOut;
	char dstFile[MAX_PATH], *data;
	int i;
	DWORD len;

	strcpy(dstFile, gSourceFile);
	i = (int) strlen(gSourceFile) - 1;	// point to last character
	for(; i > 0; i--)
	{
		if(dstFile[i] == '.')	// find extension
		{
			dstFile[i] = 0;	// NULL terminate
			strcat(dstFile, ".txt");	// add text extension
			i = 0;
		}
	}

	data = readDataFile(&len);
	if(data == NULL) return;

	// for the text case, we just write back the same file with a ".txt" extension
	fptrOut = fopen(dstFile, "wb");
	if(fptrOut == NULL)
	{
		fprintf(stderr, "\n\nError! Could not open output file \"%s\" for writing.\n\n", dstFile);
		return;
	}

	fwrite(data, 1, len, fptrOut);
	free(data);
	fclose(fptrOut);
	return;
} // wrapInTXT

// wraps the data into a bit map file
void wrapInBMP()
{
	char dstFile[MAX_PATH], *data;
	int i;
	DWORD len;

	strcpy(dstFile, gSourceFile);
	i = (int) strlen(gSourceFile) - 1;	// point to last character
	for(; i > 0; i--)
	{
		if(dstFile[i] == '.')	// find extension
		{
			dstFile[i] = 0;	// NULL terminate
			strcat(dstFile, ".bmp");	// add text extension
			i = 0;
		}
	}

	data = readDataFile(&len);
	if(data == NULL) return;

	write_BMP(data, len, dstFile, 0, gSpecialOption);
	free(data);
	return;
} // wrapInBMP

void wrapInWAV()
{
	char dstFile[MAX_PATH], *data;
	int i;
	DWORD len;

	strcpy(dstFile, gSourceFile);
	i = (int) strlen(gSourceFile) - 1;	// point to last character
	for(; i > 0; i--)
	{
		if(dstFile[i] == '.')	// find extension
		{
			dstFile[i] = 0;	// NULL terminate
			strcat(dstFile, ".wav");	// add text extension
			i = 0;
		}
	}

	data = readDataFile(&len);
	if(data == NULL) return;

	write_WAV(data, len, dstFile, gWrapperType); //
	free(data);
	return;
} // wrapInWAV

// prints the way to use this program
void usage()
{
	printf("\n\nWrappers_%4.2f.exe -f SourceFile -t FileType [-s SpecialOptions]\n\n", VERSION);
	printf("  SourceFile is the file to wrap.\n\n");
	printf("  FileType is one of the follwing types:\n\n");
	printf("    txt	- test file, changes extension only\n\n");
	printf("    bmp	- bitmap file - default 8-bit grayscale\n");
	printf("\t  [-s g] grayscale, [-s 8] 8-bit color, [-s 24] 24-bit color\n\n");
	printf("    wav	- wave file --> default 8-bit, mono, 44KHz\n");
	printf("\t  [-s 8] 8-bit samples, [-s 16] 16-bit samples\n\n");
	printf("    jpg	- jpeg file\n\n");
	printf("    pe	- portable executable file (exe, dll, scr, sys)\n");
	printf("\t  [-s CoverFile]  MUST choose an existing PE file to wrap it in.\n\n");

	printf("\n\nTo Unwrap Data, use the -u option.\n\n");
	printf("Wrappers -f SourceFile -u FileName\n\n");

	return;
} // usage

// determines the options in the command line
void parseCommandLine(int argc, char *argv[])
{
	char c1, c2, tmp[1024];
	int cnt, specialCnt;

	specialCnt = 0;	// stores special option so can fix after knowing file type
	cnt = 1;
	while(cnt < argc)
	{
		strcpy(tmp, argv[cnt]);
		c1 = *tmp;
		c2 = *(tmp + 1);

		if(c1 != '-')
		{
			usage();
			exit(0);
		}

		cnt++;	// get to next argument
		switch(c2)
		{
		case 'f': // filename
			strcpy(gSourceFile, argv[cnt]);
			break;

		case 't':		// file type
			if(stricmp(argv[cnt], "txt") == 0)
				gWrapperType = TXT;
			else if(stricmp(argv[cnt], "bmp") == 0)
				gWrapperType = BMP;
			else if(stricmp(argv[cnt], "wav") == 0)
				gWrapperType = WAV;
			else if(stricmp(argv[cnt], "PE") == 0)
				gWrapperType = PE;
			else if(stricmp(argv[cnt], "JPG") == 0)
				gWrapperType = JPG;
			break;

		case 's':		// special options
			specialCnt = cnt;
			break;

		case 'u':
			strcpy(gDestFile, argv[cnt]);
			break;

		default:	// error
			usage();
			break;
		}
		cnt++;
	} // while

	// check for special options
	if(specialCnt != 0)
	{
		if(gWrapperType == PE)
		{
			strcpy(gCoverFile, argv[specialCnt]);	// cover file name
		}
		else
		{
			gSpecialOption = *argv[specialCnt];
		}
	} // end check for special options
	return;
} // parseCommandLine

void getWrapperType(char *srcFile)
{
	// get the file extension of the source file
	gWrapperType = TXT;
	return;
} // getWrapperType

// main program to execute wrappers
void main(int argc, char *argv[])
{
	printf("\n\n Wrappers  Version: %4.2f\n\n", VERSION);
	printf("    by John A. Ortiz, last update Nov 2011\n\n");

	if(argc < 3)
	{
		usage();
		return;
	}

	// Global Variables set by parse command line
	gSourceFile[0] = 0;	// NULL terminate
	gDestFile[0] = 0;
	gWrapperType = 0;
	gSpecialOption = 0;
	gCoverFile[0] = 0;
	parseCommandLine(argc, argv);

	// make sure proper options are set
	if(gSourceFile[0] == 0)
	{
		usage();
		return;
	}

	if(gDestFile[0] != 0)	// unwrap, get wrapper type from extension
	{
		getWrapperType(gSourceFile); // sets gWrapperType
	}

	if(gWrapperType == 0)
	{
		usage();
		return;
	}

	// determine the type of wrapper
	switch(gWrapperType)
	{
	case TXT:
		wrapInTXT();
		break;

	case JPG:
		printf("\n\nWrapping in JPG not yet implemented.\n\n");
		break;

	case WAV:
		wrapInWAV();
		break;

	case BMP:
		wrapInBMP();
		break;

	case PE:
		printf("\n\nWrapping in PE not yet implemented.\n\n");
		break;
	}

	return;
} // main
