// WriteBMP.cpp
//
// This program makes a bitmap file based on an input file's size
//
// this program will automatically set
// the required vertical resolution
//
// Bitmaps are formed from the lower left, so the timeline begins in 
// the lower left and rises vertically
//
// It also makes a histogram of the data as a seperate BMP file
// 

#include "Wrappers.h"

// intialize global variables for writing the bitmap header
void initBMP_Header(char type)
{
	memset(&gBM_FileHeader, 0 , sizeof(gBM_FileHeader) );
	memset(&gBM_InfoHeader, 0 , sizeof(gBM_InfoHeader) );

	gBM_FileHeader.bfType = 0x4D42;	// "BM"
	gBM_FileHeader.bfSize = 0;			// determined later #bytes in file
	gBM_FileHeader.bfReserved1 = 0;
	gBM_FileHeader.bfReserved2 = 0;

	if(type == 'g' || type == '8')	// for 8-bit grayscale or color
	{
		gBM_FileHeader.bfOffBits = 0x436;
		gBM_InfoHeader.biBitCount = 8;		// 8 bits/pixel
		gBM_InfoHeader.biClrUsed = 256;		// 0 seems to work in 8-bit color images
	}
	else // if(type = '2')	
	{
		gBM_FileHeader.bfOffBits = 0x36;	// 24-bit color
		gBM_InfoHeader.biBitCount = 24;		// 24 bits/pixel
		gBM_InfoHeader.biClrUsed = 0;		// 24-bit color
	}

	gBM_InfoHeader.biSize = 40;
	gBM_InfoHeader.biWidth = 0;		// determined later #horizontal pixels
	gBM_InfoHeader.biHeight = 0;	// determined later #vertical pixels
	gBM_InfoHeader.biPlanes = 1;

	gBM_InfoHeader.biCompression = 0;
	gBM_InfoHeader.biSizeImage = 0;	// determined later #bytes in image portion
	gBM_InfoHeader.biXPelsPerMeter = 0;
	gBM_InfoHeader.biYPelsPerMeter = 0;
	gBM_InfoHeader.biClrImportant =0;

	return;
} // initBMP_Header

// this file writes the bmp header data
int writeBMP_Hdr(FILE *fptr, DWORD dataLen, long horizontalWidth, char type)
{
	DWORD numBytesInImage;
	long verticalHeight;
	int i, numPadBytes, padBytesLine, bytesPerPixel, bytesPerLine;
	BYTE buf[64];

	bytesPerPixel = 1;
	if(gBM_InfoHeader.biBitCount == 24) bytesPerPixel = 3;

printf("bpp=%d\n",bytesPerPixel );

	i = horizontalWidth % 4;
	padBytesLine = 0;
	if(i != 0)	padBytesLine = 4 - i; // bytes/line to pad with

	bytesPerLine = horizontalWidth * bytesPerPixel + padBytesLine;

	i = 0;
	if( dataLen % bytesPerLine != 0) i = 1;
	verticalHeight = dataLen/bytesPerLine + i;

	numBytesInImage = bytesPerLine * verticalHeight * bytesPerPixel;

printf("hwidth=%d, vheight=%d, b_per_line=%d, imagesize = %d\n", horizontalWidth, verticalHeight, bytesPerLine, numBytesInImage);
printf("hdr size=%d\n", gBM_FileHeader.bfOffBits);


	gBM_FileHeader.bfSize = numBytesInImage + gBM_FileHeader.bfOffBits;	// #bytes in file
	gBM_InfoHeader.biWidth = horizontalWidth;	// #horizontal pixels
	gBM_InfoHeader.biHeight = verticalHeight;	// #vertical pixels
	gBM_InfoHeader.biSizeImage = numBytesInImage;	// #bytes in image portion

	memset(buf, 0, 64);
	memcpy(buf, &gBM_FileHeader, sizeof(BITMAPFILEHEADER) );
	memcpy(&buf[sizeof(BITMAPFILEHEADER)], &gBM_InfoHeader, sizeof(BITMAPINFOHEADER) );
	
	fwrite(buf, 1, sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER), fptr);

	// for 24 bit color, there is no palette
	if(type == '2') return(padBytesLine);

	if(type == 'g')
	{
		// builds the gray scale palette
		for(i = 0; i < 256; i++)
		{
			fputc(i, fptr);
			fputc(i, fptr);
			fputc(i, fptr);
			fputc(0, fptr);
		}
	}
	else
	{
		// builds a standard 8-bit color palette
		for(i = 0; i < 256; i++)
		{
			int r, g, b;

			r = gStandardPalette[4 * i];
			g = gStandardPalette[4 * i + 1];
			b = gStandardPalette[4 * i + 2];

			fputc(r, fptr);
			fputc(g, fptr);
			fputc(b, fptr);
			fputc(0, fptr);
		}
	}

	return(padBytesLine);	// data offset for 8-bit gray scale bitmaps
} // writeBMP_Hdr

// opens specific BMP file to store the data
void write_BMP(char *data, DWORD dataLen, char *bmpFile, int hWidth, char type)
{
	FILE *fptr1;
	char ch1;
	int cnt, padBytesLine, tmpi;
	DWORD i;

	fptr1 = fopen(bmpFile, "wb+");
	if(fptr1 == NULL)
	{
		fprintf(stderr, "\n\nError! Could not open %s for writing.\n\n", bmpFile);
		return;
	}

	// initialize the bitmap header
	initBMP_Header(type);	// '8'==8-bit color, or '2'==24-bit

	// width in pixels
	if(hWidth == 0)
	{
		i = dataLen;
		if(type == '2') i = i / 3;
		hWidth = (int) sqrt((double) i) + 1;	// width in PIXELS
	}
	padBytesLine = writeBMP_Hdr(fptr1, dataLen, hWidth, type);

	cnt = 0;	// used for putting the required padding in BMP files
	for(i = 0; i < dataLen; i++)
	{
		int x;
		ch1 = *(data + i);

		x = fputc(ch1, fptr1);
		if(x == EOF)
		{
			// printf("len = %d, i=%d, ch1=%02x, x= %02x\n", len, i, ch1, x);
			break;
		}

	} // for

	// write zeros for any remianing bytes
	i += gBM_FileHeader.bfOffBits;	// already wrote the header
printf("data len = %d, i = %d, size=%d\n", dataLen, i, gBM_FileHeader.bfSize);
	for( ;i < gBM_FileHeader.bfSize; i++)
	{
		fputc(0, fptr1);
	}
	fclose(fptr1);
	return;
} // write_BMP

