// main.cpp
//
// this program reads a wave file and displays the header info
//

#include "Wrappers.h"

typedef struct
{
	DWORD	chunkID;
	DWORD	chunkSize;
} W_CHUNK;

typedef struct
{
	WORD	compCode;				
    WORD    numChannels;         
    DWORD   sampleRate;    
    DWORD   avgBytesPerSec;   //  avgBytesPerSec = sampleRate * blockAlign 
    WORD    blockAlign;       //  blockAlign = bitsPerSample / 8 * numChannels
    WORD    bitsPerSample;    
	// WORD extraFormat;		// do not worry about this
} W_FORMAT;


W_CHUNK gfChunk[4];	// chunk headers for the wave file
W_FORMAT gfFormatChunk;

/*
// reads in the data portion of a chunk
BYTE *readChunkData(FILE *fptr, int size)
{
	BYTE *ptr;
	int tmp, x;

	tmp = size%2;	// size MUST be WORD aligned
	if(tmp != 0) size = size + 1;

	ptr = (BYTE *) malloc(size);


	x = (int) fread(ptr, 1, size, fptr);
	
} // readChunkData
//*/

// initialize the values in the header that define the data format
void initWAV_Header(DWORD dataLen, char type)
{
	gfChunk[0].chunkID = 0x46464952;	// "RIFF"
	//filesize minus 8, RIFF,size,WAVE,fmt ,size,16 bytes,data,size
	gfChunk[0].chunkSize = dataLen + 44 - 8;

	gfChunk[1].chunkID = 0x20746d66;	// "fmt "
	gfChunk[1].chunkSize = 16;

	// set values for wav file
	gfFormatChunk.compCode = 1;	// (2)
	gfFormatChunk.numChannels = 1;	// (2) mono vs. stereo

	if(gSpecialOption == '8')
        gfFormatChunk.bitsPerSample = 8; //  (2)
	else if(gSpecialOption == '1')	// assuming 16-bits per sample
		gfFormatChunk.bitsPerSample = 16;

	gfFormatChunk.sampleRate = 44100;	// (4)
	gfFormatChunk.blockAlign = (gfFormatChunk.bitsPerSample / 8) * gfFormatChunk.numChannels; // (2)
	gfFormatChunk.avgBytesPerSec = gfFormatChunk.sampleRate * gfFormatChunk.blockAlign; // (4)

	gfChunk[2].chunkID = 0x61746164;	// "data"
	gfChunk[2].chunkSize = dataLen;
	return;
} // initWAV_Header

// write the header of the wave file
void writeWAV_Hdr(FILE *fptr, DWORD dataLen, char type)
{
	char wString[] = "WAVE";

	fwrite(&gfChunk[0].chunkID, 1, 4, fptr);
	fwrite(&gfChunk[0].chunkSize, 1, 4, fptr);
	fwrite(wString, 1, 4, fptr);
	fwrite(&gfChunk[1].chunkID, 1, 4, fptr);
	fwrite(&gfChunk[1].chunkSize, 1, 4, fptr);

	//WORD	compCode;				
 //   WORD    numChannels;         
 //   DWORD   sampleRate;    
 //   DWORD   avgBytesPerSec;   //  avgBytesPerSec = sampleRate * blockAlign 
 //   WORD    blockAlign;       //  blockAlign = bitsPerSample / 8 * numChannels
 //   WORD    bitsPerSample;    

	fwrite(&gfFormatChunk.compCode, 1, 2, fptr);
	fwrite(&gfFormatChunk.numChannels, 1, 2, fptr);
	fwrite(&gfFormatChunk.sampleRate, 1, 4, fptr);
	fwrite(&gfFormatChunk.avgBytesPerSec, 1, 4, fptr);
	fwrite(&gfFormatChunk.blockAlign, 1, 2, fptr);
	fwrite(&gfFormatChunk.bitsPerSample, 1, 2, fptr);

	// write the data
	fwrite(&gfChunk[2].chunkID, 1, 4, fptr);
	fwrite(&gfChunk[2].chunkSize, 1, 4, fptr);


	return;
} // writeWAV_Hdr

// writes the wave data
void write_WAV(char *data, DWORD dataLen, char *wavFile, char type)
{
	FILE *fptr1;

	fptr1 = fopen(wavFile, "wb+");
	if(fptr1 == NULL)
	{
		fprintf(stderr, "\n\nError! Could not open %s for writing.\n\n", wavFile);
		return;
	}

	// initialize the header
	initWAV_Header(dataLen, type);
	writeWAV_Hdr(fptr1, dataLen, type);
	fwrite(data, 1, dataLen, fptr1);	// write the rest of the data

	fclose(fptr1);
	return;
} // write_WAV


// prints out wave format info
//void printFormat(W_FORMAT fmt)
//{
//	printf("\n\nWAVE FORMAT INFO\n\n");
//	printf("Compression Code:		%d\n", fmt.compCode);
//	printf("Number of Channels:		%d\n", fmt.numChannels);
//	printf("Sample Rate: 			%d\n", fmt.sampleRate);
//	printf("Average Bytes/Second:		%d\n", fmt.avgBytesPerSec);
//	printf("Block Align: 			%d\n", fmt.blockAlign);
//	printf("Bits per Sample: 		%d\n", fmt.bitsPerSample);
//	return;
//} // printFormat


