// WriteBMP.cpp
//
// This program makes a bitmap file based on an input file's size
//
// The user specifies the horizontal dimension, and based upon
// the input file's size, this program will automatically set
// the required vertical resolution
//
// Bitmaps are formed from the lower left, so the timeline begins in 
// the lower left and rises vertically
//
// It also makes a histogram of the data as a seperate BMP file
// 

#include <windows.h>
#include <stdio.h>
#include <io.h>
#include <math.h>
#include "WriteBMP.h"

// Global Veriable Definitions

BYTE gFileReadBuffer[32 * 1024 * 1024];	// 32 MB will be sufficient in most cases
RGB gPalette[256];

BITMAPFILEHEADER gBM_FileHeader;
BITMAPINFOHEADER gBM_InfoHeader;

QWORD gHistogram[256];
QWORD gSortedHistogram[256];
QWORD gBitHistogram[2];
QWORD gTmpHistogram[256];
QWORD gHistogram2D[256][256];

DWORD ghWidth = 0;
float gZoomFactor = 1.0;
bool gPercentFlag = false;
bool gLogFlag = false;
bool gScaledFlag = false;
bool gZoomTop = false;
bool g2Dflag = false;
bool gSortFlag = false;

bool gCreateBMP = false;	// when true, will create BMP image of target file

// for histogram purposes, this table holds the number of "1" bits for each hex value
char gBitCount[256] = 
{
	0, 1, 1, 2, 1, 2, 2, 3,
	1, 2, 2, 3, 2, 3, 3, 4,
	1, 2, 2, 3, 2, 3, 3, 4,
	2, 3, 3, 4, 3, 4, 4, 5,

	1, 2, 2, 3, 2, 3, 3, 4,
	2, 3, 4, 4, 3, 4, 4, 5,
	2, 3, 3, 4, 3, 4, 4, 5,
	3, 4, 4, 5, 4, 5, 5, 6,

	1, 2, 2, 3, 2, 3, 3, 4,
	2, 3, 3, 4, 3, 4, 4, 5,
	2, 3, 3, 4, 3, 4, 4, 5,
	3, 4, 4, 5, 4, 5, 5, 6,

	2, 3, 3, 4, 3, 4, 4, 5,
	3, 4, 4, 5, 4, 5, 5, 6,
	3, 4, 4, 5, 4, 5, 5, 6,
	4, 5, 5, 6, 5, 6, 6, 7,

	// > 128

	1, 2, 2, 3, 2, 3, 3, 4,
	2, 3, 3, 4, 3, 4, 4, 5,
	2, 3, 3, 4, 3, 4, 4, 5,
	3, 4, 4, 5, 4, 5, 5, 6,

	2, 3, 3, 4, 3, 4, 4, 5,
	3, 4, 4, 5, 4, 5, 5, 6,
	3, 4, 4, 5, 4, 5, 5, 6,
	4, 5, 5, 6, 5, 6, 6, 7,

	2, 3, 3, 4, 3, 4, 4, 5,
	3, 4, 4, 5, 4, 5, 5, 6,
	3, 4, 4, 5, 4, 5, 5, 6,
	4, 5, 5, 6, 5, 6, 6, 7,

	3, 4, 4, 5, 4, 5, 5, 6,
	4, 5, 5, 6, 5, 6, 6, 7,
	4, 5, 5, 6, 5, 6, 6, 7,
	5, 6, 6, 7, 6, 7, 7, 8
};

// intialize global variables for writing the bitmap header
void initBM_Header()
{
	gBM_FileHeader.bfType = 0x4D42;	// "BM"
	gBM_FileHeader.bfSize = 0;			// determined later #bytes in file
	gBM_FileHeader.bfReserved1 = 0;
	gBM_FileHeader.bfReserved2 = 0;
	gBM_FileHeader.bfOffBits = 0x436;	// for 8-bit grayscale

	gBM_InfoHeader.biSize = 40;
	gBM_InfoHeader.biWidth = 0;		// determined later #horizontal pixels
	gBM_InfoHeader.biHeight = 0;	// determined later #vertical pixels
	gBM_InfoHeader.biPlanes = 1;
	gBM_InfoHeader.biBitCount = 8;	// 8 bits/pixel
	gBM_InfoHeader.biCompression = 0;
	gBM_InfoHeader.biSizeImage = 0;	// determined later #bytes in image portion
	gBM_InfoHeader.biXPelsPerMeter = 0;
	gBM_InfoHeader.biYPelsPerMeter = 0;
	gBM_InfoHeader.biClrUsed = 256;	// 8-bit grayscale
	gBM_InfoHeader.biClrImportant =0;

	return;
} // initBM_Header

double calculateBitEntropy()
{
	QWORD maxCount, sum;
	double prb0, prb1, bitEntropy, lg2;

	lg2 = (double) log( (double) 2.00);	// just calculate this once

	maxCount = gBitHistogram[0];
	if( gBitHistogram[1] > gBitHistogram[0] ) maxCount = gBitHistogram[1];
	sum = gBitHistogram[0] + gBitHistogram[1];

	prb0 = (double) gBitHistogram[0]/sum;
	prb1 = (double) gBitHistogram[1]/sum;

	bitEntropy = 0;
	if(prb0 != 0.0)
	{
		bitEntropy += -1.0 * prb0 * ( log(prb0) / lg2 );
	}

	if(prb1 != 0.0)
	{
		bitEntropy += -1.0 * prb1 * ( log(prb1) / lg2 );
	}

	return bitEntropy;
} // calculateBitEntropy

double calculateEntropy(QWORD byteCount)
{
	double Entropy, sum, lg2, prob;

	lg2 = (double) log( (double) 2.00);	// just calculate this once

	Entropy = 0;.0;
	for(int i = 0; i < 256; i++)
	{
		prob = (double) gHistogram[i] / byteCount;

		if(prob != 0.0)
		{
			Entropy += prob * log( 1 / prob ) / lg2;
		}
	}

	return Entropy;
} // calculateEntropy

// this file writes the standard grayscale bmp header data
int write_8bitHdr(FILE *fptr, DWORD sizeInputFile, long horizontalWidth, RGB palette[256])
{
	DWORD numBytesInImage;
	long verticalHeight;
	int i, roundup, numPadBytes, padBytesLine;
	BYTE buf[64];

	roundup = 1;
	if( (sizeInputFile % horizontalWidth) == 0)
		roundup = 0;
	verticalHeight = sizeInputFile/horizontalWidth + roundup;
	i = horizontalWidth % 4;
	padBytesLine = 0;
	if(i != 0)	padBytesLine = 4 - i; // bytes/line to pad with

	numPadBytes = verticalHeight * padBytesLine; // total bytes to pad
	numBytesInImage = horizontalWidth * verticalHeight + numPadBytes;

	gBM_FileHeader.bfSize = numBytesInImage + 0x436;		// #bytes in file
	gBM_InfoHeader.biWidth = horizontalWidth;	// #horizontal pixels
	gBM_InfoHeader.biHeight = verticalHeight;	// #vertical pixels
	gBM_InfoHeader.biSizeImage = numBytesInImage;	// #bytes in image portion

	memcpy(buf, &gBM_FileHeader, sizeof(BITMAPFILEHEADER) );
	memcpy(&buf[sizeof(BITMAPFILEHEADER)], &gBM_InfoHeader, sizeof(BITMAPINFOHEADER) );
	
	fwrite(buf, 1, sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER), fptr);

	// write the palette
	for(i = 0; i < 256; i++)
	{
		fputc(palette[i].blu, fptr);
		fputc(palette[i].grn, fptr);
		fputc(palette[i].red, fptr);
		fputc(0, fptr);
	}

	return(padBytesLine);	// data offset for 8-bit gray scale bitmaps
} // write_8bitHdr

// this file writes the standard grayscale bmp header data
int write_8bitGrayHdr(FILE *fptr, DWORD sizeInputFile, long horizontalWidth)
{
	DWORD numBytesInImage;
	long verticalHeight;
	int i, roundup, numPadBytes, padBytesLine;
	BYTE buf[64];

	roundup = 1;
	if( (sizeInputFile % horizontalWidth) == 0)
		roundup = 0;
	verticalHeight = sizeInputFile/horizontalWidth + roundup;
	i = horizontalWidth % 4;
	padBytesLine = 0;
	if(i != 0)	padBytesLine = 4 - i; // bytes/line to pad with

	numPadBytes = verticalHeight * padBytesLine; // total bytes to pad
	numBytesInImage = horizontalWidth * verticalHeight + numPadBytes;

	gBM_FileHeader.bfSize = numBytesInImage + 0x436;		// #bytes in file
	gBM_InfoHeader.biWidth = horizontalWidth;	// #horizontal pixels
	gBM_InfoHeader.biHeight = verticalHeight;	// #vertical pixels
	gBM_InfoHeader.biSizeImage = numBytesInImage;	// #bytes in image portion

	memcpy(buf, &gBM_FileHeader, sizeof(BITMAPFILEHEADER) );
	memcpy(&buf[sizeof(BITMAPFILEHEADER)], &gBM_InfoHeader, sizeof(BITMAPINFOHEADER) );
	
	fwrite(buf, 1, sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER), fptr);

	// write the palette
	for(i = 0; i < 256; i++)
	{
		fputc(i, fptr);
		fputc(i, fptr);
		fputc(i, fptr);
		fputc(0, fptr);
	}

	return(padBytesLine);	// data offset for 8-bit gray scale bitmaps
} // write_8bitGrayHdr


// sorts a histogram in ascending order
void getSortedHistogram(QWORD histogram[], QWORD sortedHistogram[])
{
	int i, j, numValuesSorted, pos;

	// sort the input histogram
	numValuesSorted = 1;
	sortedHistogram[0] = histogram[0];	// initial value
	for(i = 1; i < 256; i++)
	{
		if(histogram[i] <= sortedHistogram[numValuesSorted-1])	// goes at end of list
		{
			sortedHistogram[numValuesSorted] = histogram[i];
			numValuesSorted++;
			continue;
		}

		// find position
		pos = numValuesSorted;
		for(j = numValuesSorted - 1; j >= 0; j--)
		{
			if(histogram[i] > sortedHistogram[j])
				pos = j;
			else
				break;
		}

		// shift the list to make room for this element
		for(j = numValuesSorted; j > pos; j--)
		{
			sortedHistogram[j] = sortedHistogram[j - 1];
		}
		sortedHistogram[pos] = histogram[i];
		numValuesSorted++;
	}

	return;
} // getSortedHistogram

DWORD getHistogram(char *binFile, QWORD *byteCount, QWORD *minCount, QWORD *min2DCnt, QWORD *maxCount, QWORD *max2DCnt, bool flag2D)
{
	FILE *fptr2;
	BYTE ch1, ch2;

	DWORD numBytesRead, x, y;
	HANDLE hFile;
	BOOL result;

	// initialize histogram
	for(x = 0; x < 256; x++)
	{
		gHistogram[x] = 0;
	}

	// initialize BIT histogram
	gBitHistogram[0] = 0;
	gBitHistogram[1] = 0;

	// initialize 2D histogram
	if(flag2D)
		for(x = 0; x < 256; x++)
			for(y = 0; y < 256; y++)
				gHistogram2D[x][y] = 0;

	hFile = CreateFile(binFile, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
	if(hFile == INVALID_HANDLE_VALUE)	// error accessing file
	{
		printf("ERROR, could not open %s for reading.\n\n", binFile);
		exit(0);
	}

	*byteCount = 0;
	do
	{
		result = ReadFile(hFile, gFileReadBuffer, 32 * 1024 * 1024, &numBytesRead, NULL);	// attemp 32MB, will normally be enough,  * 1024 * 1024
		if(result == FALSE)	// && GetLastError() != ERROR_HANDLE_EOF)
		{
			fprintf(stderr, "\n\nError - Failed to correctly read file.\n\n");
			exit(0);
		}
		//fprintf(stderr, "Read %u bytes\n\n", numBytesRead);
		*byteCount += (QWORD) numBytesRead;

		for(x = 0; x < numBytesRead; x++)
		{
			ch1 = gFileReadBuffer[x];
			gHistogram[ch1]++;

			gBitHistogram[0] += (8 - gBitCount[ch1]);
			gBitHistogram[1] += gBitCount[ch1];			// reference table that has # of ones in a hex value

			if(flag2D)
				if( (x+1) < numBytesRead)
				{
					ch2 = gFileReadBuffer[x+1];
					gHistogram2D[ch1][ch2]++;
				}
		} // for
	}
	while(numBytesRead > 0);
	CloseHandle(hFile);

	*maxCount = 0;
	*minCount = 0xFFFFFFFFFFFFFFFF;	// initialize to max
	*min2DCnt = *minCount;
	*max2DCnt = 0;
	for(x = 0; x < 256; x++)
	{
		if(gHistogram[x] < *minCount) *minCount = gHistogram[x];
		if(gHistogram[x] > *maxCount) *maxCount = gHistogram[x];
	}

	// Get min/max 2D values for scaling output bitmap
	if(flag2D)
	{
		for(x = 0; x < 256; x++)
			for(y = 0; y < 256; y++)
			{
				if(gHistogram2D[x][y] < *min2DCnt) *min2DCnt = gHistogram2D[x][y];
				if(gHistogram2D[x][y] > *max2DCnt) *max2DCnt = gHistogram2D[x][y];
			}
	}
	return SUCCESS;
} // getHistogram

// creates a text-based histogram
void printTextualHistogram(FILE *fptr, int i, QWORD count, QWORD maxCount, QWORD byteCount)
{
	QWORD scale, tmp;
	char buf[116];
	double p;

	if(fptr == NULL) return;

	scale = maxCount/100 + 1;

	p =  100 * ( (double) count/byteCount);
	if(i == 9)
	{
		fprintf(fptr, "TAB %03d [%02X],%I64u (%6.3f%%)", i, i, count, p);
	}
	else if(i == 10)	// LF
	{
		fprintf(fptr, "LF, %03d [%02X],%I64u (%6.3f%%)", i, i, count, p);
	}
	else if(i == 13) // CR
	{
		fprintf(fptr, "CR, %03d [%02X],%I64u (%6.3f%%)", i, i, count, p);
	}
	else if(i == 32)
	{
		fprintf(fptr, "SP, %03d [%02X],%I64u (%6.3f%%)", i, i, count, p);
	}
	else if(i >= 28 && i <= 31)
	{
		fprintf(fptr, "  , %03d [%02X],%I64u (%6.3f%%)", i, i, count, p);
	}
	else
	{
		fprintf(fptr, "%2c, %03d [%02X],%I64u (%6.3f%%)", i, i, i, count, p);
	}

	tmp = count/scale;
	if(tmp == 0 && count > 0) tmp = 1;
	if(tmp >= 104)
	{
		fprintf(stderr, "TMP >= 104 ==> %I64u\n\n", tmp);
		fflush(stderr);
		return;
	}
	memset(buf, 0, 116);
	memset(buf, 45, (size_t) tmp);
	if(tmp >= 9) buf[9] = 43;
	if(tmp >= 19) buf[19] = 43;
	if(tmp >= 29) buf[29] = 43;
	if(tmp >= 39) buf[39] = 43;
	if(tmp >= 49) buf[49] = 43;
	if(tmp >= 59) buf[59] = 43;
	if(tmp >= 69) buf[69] = 43;
	if(tmp >= 79) buf[79] = 43;
	if(tmp >= 89) buf[89] = 43;
	if(tmp >= 99) buf[99] = 43;
	fprintf(fptr, "%s\n", buf);

	return;
} // printTextualHistogram

// print both counts
void printTextualBitHistogram(FILE *fptr)
{
	QWORD maxCount, sum, scale, tmp;
	char buf[116];
	double p0, p1;

	if(fptr == NULL) return;

	fprintf(fptr, "\n");
	maxCount = gBitHistogram[0];
	if( gBitHistogram[1] > gBitHistogram[0] ) maxCount = gBitHistogram[1];
	sum = gBitHistogram[0] + gBitHistogram[1];

	scale = maxCount/100 + 1;

	p0 =  100 * ( (double) gBitHistogram[0]/sum);	// calculate percentage
	p1 =  100 * ( (double) gBitHistogram[1]/sum);

	// print "0" count
	fprintf(fptr, "0, %I64u (%6.3f%%)", gBitHistogram[0], p0);

	tmp = gBitHistogram[0]/scale;
	if(tmp == 0 && gBitHistogram[0] > 0) tmp = 1;

	memset(buf, 0, 116);
	memset(buf, 45, (size_t) tmp);
	if(tmp >= 9) buf[9] = 43;
	if(tmp >= 19) buf[19] = 43;
	if(tmp >= 29) buf[29] = 43;
	if(tmp >= 39) buf[39] = 43;
	if(tmp >= 49) buf[49] = 43;
	if(tmp >= 59) buf[59] = 43;
	if(tmp >= 69) buf[69] = 43;
	if(tmp >= 79) buf[79] = 43;
	if(tmp >= 89) buf[89] = 43;
	if(tmp >= 99) buf[99] = 43;
	fprintf(fptr, "%s\n", buf);

	// print "1" count
	fprintf(fptr, "1, %I64u (%6.3f%%)", gBitHistogram[1], p1);
	tmp = gBitHistogram[1]/scale;
	if(tmp == 0 && gBitHistogram[1] > 0) tmp = 1;

	memset(buf, 0, 116);
	memset(buf, 45, (size_t) tmp);
	if(tmp >= 9) buf[9] = 43;
	if(tmp >= 19) buf[19] = 43;
	if(tmp >= 29) buf[29] = 43;
	if(tmp >= 39) buf[39] = 43;
	if(tmp >= 49) buf[49] = 43;
	if(tmp >= 59) buf[59] = 43;
	if(tmp >= 69) buf[69] = 43;
	if(tmp >= 79) buf[79] = 43;
	if(tmp >= 89) buf[89] = 43;
	if(tmp >= 99) buf[99] = 43;
	fprintf(fptr, "%s\n", buf);

	fprintf(fptr, "\n");
	return;
} // printTextualBitHistogram

// opens specific BMP file to store the data
void write_BMP_histogram(QWORD maxCount, FILE *fptrTxtHist, FILE *fptrHist, QWORD histogram[], QWORD byteCount, QWORD minCount)
{
	int padBytesLine;
	DWORD h, i, j;
	BYTE *ptrB, *tmpPtr, shadeOfGray;
	double tmpDD, scale;

	// already found maxCount
	if(maxCount == 0) return;

	ghWidth = BMP_WIDTH;
	// setting width to 516 pixels, results in padBytesLine == 0
	padBytesLine = write_8bitGrayHdr(fptrHist, BMP_WIDTH * BMP_HEIGHT, ghWidth); // will always return zero here, width == 516

	// allocate memory for bitmap
	ptrB = (BYTE *) malloc(BMP_WIDTH * BMP_HEIGHT);
	if(ptrB == NULL)
	{
		printf("ERROR, could not allocate memory for bit map image.\n\n");
		return;
	}
	memset(ptrB, 255, BMP_WIDTH * BMP_HEIGHT); // make background white

	// treat this memory as an array 512 * HISTOGRAM_HEIGHT
	// scale histogram to HISTOGRAM_HEIGHT - 16
	shadeOfGray = 0;
	scale = ((double) HISTOGRAM_HEIGHT)/maxCount;
	for(i = 0; i < 256; i++)
	{
//		if(fptrTxtHist != NULL) fprintf(fptrTxtHist, "%03d,%d\n", i, histogram[i]);
		printTextualHistogram(fptrTxtHist, i, histogram[i], maxCount, byteCount);	// prints only 1 character count

		tmpDD = (double) histogram[i];
		if(gZoomTop == true)
		{
			tmpDD = tmpDD - minCount;
		}
		tmpDD = tmpDD * scale;

		if(tmpDD == 0 && histogram[i] > 0) tmpDD = 1.0; // want at least one black dot if not zero

		// have BM height=532, Hist height=512, leaves 20 lines @ 516 pixels each
		tmpPtr = ptrB +		// horizontal white line padding					vertical column padding
			( (BMP_HEIGHT - HIST_HEIGHT)/2 * BMP_WIDTH) + i * 2 + (BMP_WIDTH - HIST_WIDTH)/2;

		if((i % 16) == 0)
			shadeOfGray = 0;	// black
		else
			shadeOfGray = 128;	// dark gray

		// draw the line upward
		h = (DWORD) (tmpDD * gZoomFactor);
		if(h > HIST_HEIGHT)
			h = HIST_HEIGHT + (BMP_HEIGHT - HIST_HEIGHT)/2;

		// put solid black line at bottom of histogram
		if(i < 255)
		{
			*(tmpPtr) = 0;
			*(tmpPtr + 1) = 0;
		}

		for(j = 1; j < h; j++)
		{
			*(tmpPtr + j * 516) = shadeOfGray;

			// i selects column
			if(i == 0 || i == 64 || i == 128 || i == 192 || i == 255)
			{
				*(tmpPtr + (j-1) * BMP_WIDTH) = 0;
				*(tmpPtr + (j-2) * BMP_WIDTH) = 0;
				*(tmpPtr + (j-3) * BMP_WIDTH) = 0;
				*(tmpPtr + (j-4) * BMP_WIDTH) = 0;
			}

			// puts delimiter marks, horizontal spacing MUST be AT LEAST 8 lines below (16 total)
			if(i == 0 || i == 128 || i == 255)
			{
				*(tmpPtr + (j-5) * BMP_WIDTH) = 0;
				*(tmpPtr + (j-6) * BMP_WIDTH) = 0;
				*(tmpPtr + (j-7) * BMP_WIDTH) = 0;
				*(tmpPtr + (j-8) * BMP_WIDTH) = 0;
			}
		}
	}
	fwrite(ptrB, 1, BMP_WIDTH * BMP_HEIGHT, fptrHist);
	free(ptrB);

	printTextualBitHistogram(fptrTxtHist);
	return;
} // write_BMP_histogram

void setPaletteColor(int index, BYTE B, BYTE G, BYTE R, RGB palette[256])
{
	palette[index].blu = B;
	palette[index].grn = G;
	palette[index].red = R;

	return;
} // setPaletteColor

void write_2D_BMP_Histogram(FILE *fptrHist, QWORD min2DCount, QWORD max2DCount, QWORD histogram2D[256][256], QWORD byteCount)
{
	int padBytesLine;
	DWORD h, i, j;
	BYTE *ptrB, *tmpPtr, shadeOfGray;
	double tmpDD, scale;

	ghWidth = 512;

	// initialize the palette
	for(i = 0; i < 256; i++)
	{
		gPalette[i].blu = i;
		gPalette[i].grn = i;
		gPalette[i].red = i;
	}

	setPaletteColor(0, 128, 128, 128, gPalette);	// index 0 = grey, background
	setPaletteColor(1, 255, 0, 0, gPalette);	// index 1 = blu
	setPaletteColor(2, 0, 255, 0, gPalette);	// index 2 = grn
	setPaletteColor(3, 0, 255, 255, gPalette);	// index 3 = yel
	setPaletteColor(4, 0, 0, 255, gPalette);	// index 4 = red

	// setting width to 516 pixels, results in padBytesLine == 0
	padBytesLine = write_8bitHdr(fptrHist, 512 * 512, 512, gPalette); // will always return zero here, width == 512

	// allocate memory for bitmap
	ptrB = (BYTE *) malloc(512 * 512);
	if(ptrB == NULL)
	{
		printf("ERROR, could not allocate memory for bit map image.\n\n");
		return;
	}

	// treat this memory as an array 512 * 512
	shadeOfGray = 0;
	scale = (double) 255 / max2DCount;	// 256 shades of gray

//fprintf(stderr, "max2D=%llu, scale=%f\n", max2DCount, scale);

	for(i = 255; i < 256; i--)	// bitmaps display upside down (i < 256, when i=0 and is decremented, it becomes 4 billion)
		for(j = 0; j < 256; j++)
		{
			BYTE value;
			int index;

			tmpDD = (double) histogram2D[i][j];
			tmpDD = tmpDD * scale;

//fprintf(stderr, "i=%u, j=%u, val=%llu, tmpDD=%u, %f\n", i, j, histogram2D[i][j], (BYTE) tmpDD, tmpDD);

			// set the pixels according to scaled count
			index = i * 2 * 512 + j * 2;

			value = 0;
			// tmpD == 0, color is gray
			if(tmpDD > 0.0) value = 1;		// index into palette, blu
			if(tmpDD > 64.0) value = 2;		// grn
			if(tmpDD > 128.0) value = 3;	// yel
			if(tmpDD > 192.0) value = 4;	// red

			*(ptrB + index) = value;
			*(ptrB + index + 1) = value;
			index += 512;
			*(ptrB + index) = value;
			*(ptrB + index + 1) = value;
		}// end nested for i & j

	fwrite(ptrB, 1, 512 * 512, fptrHist);
	free(ptrB);
	return;
} // write_2D_BMP_Histogram

// writes a histogram for every option
void writeHistograms(char *binFile)
{
	char tmpFileName[MAX_PATH], histFile[MAX_PATH], tmpBuf[32];
	FILE *fptrTxtHist, *fptrHist;
	QWORD max, minCount, maxCount, byteCount, min2DCount, max2DCount;
	double Entropy, bitEntropy;
	DWORD x;

	x = getHistogram(binFile, &byteCount, &minCount, &min2DCount, &maxCount, &max2DCount, g2Dflag);
	if(x != SUCCESS) return;

	Entropy = calculateEntropy(byteCount);
	bitEntropy = calculateBitEntropy();

	// open the bitmap output file
	sprintf(tmpFileName, "hist_%s_z_%03d.bmp", binFile, (int) gZoomFactor);

	fptrHist = fopen(tmpFileName, "wb+");
	if(fptrHist == NULL)
	{
		printf("ERROR, could not open %s for writing.\n\n", tmpFileName);
		return;
	}

	sprintf(histFile, "hist_%s.txt", binFile);
	fptrTxtHist = fopen(histFile, "w+");
	if(fptrTxtHist == NULL)
	{
		fclose(fptrHist);
		printf("Could not open text histogram file for writing.\n\n");
		return;
	}

	write_BMP_histogram(maxCount, fptrTxtHist, fptrHist, gHistogram, byteCount, minCount);
	fprintf(stdout, "\nFile Entropy = %7.5f\n\n", Entropy);
	fprintf(fptrTxtHist, "\nFile Entropy = %7.5f\n\n", Entropy);
	fprintf(stdout, "\nBit Entropy = %7.5f\n\n", bitEntropy);
	fprintf(fptrTxtHist, "\nBit Entropy = %7.5f\n\n", bitEntropy);
	fclose(fptrTxtHist);
	fclose(fptrHist);
	fptrTxtHist = NULL;

	if(gSortFlag)
	{
		getSortedHistogram(gHistogram, gSortedHistogram);

		// open the bitmap output file
		sprintf(tmpFileName, "hist_sort_%s_z_%03d.bmp", binFile, (int) gZoomFactor);

		fptrHist = fopen(tmpFileName, "wb+");
		if(fptrHist == NULL)
		{
			printf("ERROR, could not open %s for writing.\n\n", tmpFileName);
			return;
		}

		sprintf(histFile, "hist_sort_%s.txt", binFile);
		fptrTxtHist = fopen(histFile, "w+");
		if(fptrTxtHist == NULL)
		{
			fclose(fptrHist);
			printf("Could not open sorted text histogram file for writing.\n\n");
			return;
		}

		write_BMP_histogram(maxCount, fptrTxtHist, fptrHist, gSortedHistogram, byteCount, minCount);
		fprintf(fptrTxtHist, "\nFile Entropy = %7.5f\n\n", Entropy);
		fprintf(fptrTxtHist, "\nBit Entropy = %7.5f\n\n", bitEntropy);
		fclose(fptrTxtHist);
		fclose(fptrHist);
		fptrTxtHist = NULL;
	}

	// open the 2D bitmap output file
	if(g2Dflag)
	{
		sprintf(tmpFileName, "hist_2D_%s_z_%03d.bmp", binFile, (int) gZoomFactor);
		fptrHist = fopen(tmpFileName, "wb+");
		if(fptrHist == NULL)
		{
			printf("ERROR, could not open %s for writing.\n\n", tmpFileName);
			return;
		}
		write_2D_BMP_Histogram(fptrHist, min2DCount, max2DCount, gHistogram2D, byteCount);
		fclose(fptrHist);
	}

	// make a logarithmic histogram if the option set
	if(gLogFlag == true)
	{
		max = 0;
		for(x = 0; x < 256; x++)
		{
			if(gHistogram[x] != 0)
				gTmpHistogram[x] =  (DWORD) ((double) log( (double) gHistogram[x]) * 10000);
			else
				gTmpHistogram[x] = 0;

			if(gTmpHistogram[x] > max) max = gTmpHistogram[x];
		}
		sprintf(tmpFileName, "hist_log_%s.bmp", binFile);
		fptrHist = fopen(tmpFileName, "wb+");
		if(fptrHist == NULL)
		{
			printf("ERROR, could not open %s for writing.\n\n", tmpFileName);
			return;
		}
		write_BMP_histogram(max, NULL, fptrHist, gTmpHistogram, byteCount, minCount);
		fclose(fptrHist);
	}

	// make a scaled histogram if the option set
	if(gScaledFlag == true)
	{
		DWORD max2, indx;

		max = 0;
		max2 = 0;
		indx = 0;
		for(x = 0; x < 256; x++)
		{
			gTmpHistogram[x] = gHistogram[x];

			if(gTmpHistogram[x] > max)
			{
				max = gTmpHistogram[x];
				indx = x;	// keep index of maximum count
			}
			else if(gTmpHistogram[x] > max2)
			{
				max2 = gTmpHistogram[x];
			}
		}
//printf("max=%d, max2=%d, indx = %d\n", max, max2, indx);
		max2 = DWORD ((double) max2 * 1.25);
		if(max > max2 && max2 != 0) 
		{
			gTmpHistogram[indx] = max2; // max value will never exceed 20%
			max = max2;
		}

		sprintf(tmpFileName, "hist_sc_%s_z_%03d.bmp", binFile, (int) gZoomFactor);
		fptrHist = fopen(tmpFileName, "wb+");
		if(fptrHist == NULL)
		{
			printf("ERROR, could not open %s for writing.\n\n", tmpFileName);
			return;
		}
		write_BMP_histogram(max, NULL, fptrHist, gTmpHistogram, byteCount, minCount);
		fclose(fptrHist);
	} // gScaled Flag

	return;
} // writeHistograms

// opens specific BMP file to store the data
void write_BMP(char *binFile)
{
	FILE *fptr1, *fptr2;
	char ch1;
	int cnt, padBytesLine, tmpLength, length, i, bmpCnt;
	char tmpBMPName[MAX_PATH];

	fptr2 = fopen(binFile, "rb");
	if(fptr2 == NULL)
	{
		printf("ERROR, could not open %s for reading.\n\n", binFile);
		return;
	}
	length = (int) filelength(_fileno(fptr2));
	tmpLength = length;
	bmpCnt = 1;	// number of the bitmap output file
	fptr1 = NULL;

	while(tmpLength > 0)
	{
		if(fptr1 != NULL) // close when writing mutliple files
		{
			fclose(fptr1);
			fptr1 = NULL;
		}

		// get a new file name
		sprintf(tmpBMPName, "%s_%03d.bmp", binFile, bmpCnt++);
		fptr1 = fopen(tmpBMPName, "wb+");
		if(fptr1 == NULL)
		{
			printf("ERROR, could not open %s for writing.\n\n", tmpBMPName);
			return;
		}

		// sets the length for each iteration
		length = tmpLength;

		// want to limit each bmp file size to MAX_BMP_SIZE
		if(length > MAX_BMP_SIZE) length = MAX_BMP_SIZE;

		if(ghWidth == 0)
		{
			ghWidth = (int) sqrt((double) length);
		}
		padBytesLine = write_8bitGrayHdr(fptr1, length, ghWidth);

		cnt = 0;	// used for putting the required padding in BMP files
		for(i = 0; i < length; i++)
		{
			int x;
			ch1 = fgetc(fptr2);

			x = fputc(ch1, fptr1);
			if(x == EOF)
			{
				printf("length = %d, i=%d, ch1=%02x, x= %02x\n", length, i, ch1, x);
				break;
			}
			// ONLY APPLIES TO LENGTHS NOT DIVISIBLE EVENLY BY 4
			cnt++;
			if(cnt == ghWidth)
			{
				cnt = padBytesLine;
				while(cnt > 0)
				{
					fputc(0, fptr1);
					cnt--;
				}
				// cnt will be set to zero here
			} // make zeros for padding bytes
		} // for i < length
		// write zeros for remaining bytes
		for( ;i < (int) gBM_FileHeader.bfSize; i++)
		{
			fputc(0, fptr1);
		}
		tmpLength -= MAX_BMP_SIZE;
	} // while tmpLength > 0

	fclose(fptr1);
	fclose(fptr2);
	return;
} // write_BMP

// shows correct usage format
void usage()
{
	printf("\n  wbh_%4.2f.exe InputFile [options]\n", VERSION);
	printf("\nThis program creates a histogram of the input file.\n");
	printf("The input file may be of ANY file type.\n\n");

	printf("   -b				Create a bitmap image of the file\n");
	printf("   -h [width]		Specifiy a horizonal dimension in pixels\n");
	printf("   -s				Scaled the histogram\n");
	printf("   -z [1 to 100]	Zoom Factor to zoom to bototm portion of histogram\n");
	printf("   -t				Zoom to top part of histogram\n");
	printf("   -sort			Create a sorted histogram\n");
//	printf("You can optionally specifiy a logarithmic histogram. (-l)\n");
//	printf("You can optionally specifiy a percentage histogram. (-p)\n");
	printf("\n");
	return;
} // usage

// set flags for operation
int parseCommandLine(int argc, char *argv[])
{
	int i;

	if(argc < 2)
	{
		usage();
		return(FAILURE);
	}

	i = 2;
	while(i < argc)
	{
		// adjust the width
		if(stricmp(argv[i], "-h") == 0)
		{
			i++;
			ghWidth = atoi(argv[i]);
		}

		// logarithmic histogram
		if(stricmp(argv[i], "-l") == 0)
		{
			gLogFlag = true;
		}

		if(stricmp(argv[i], "-2d") == 0)
		{
			g2Dflag = true;
		}

		// bitmap image
		if(stricmp(argv[i], "-b") == 0)
		{
			gCreateBMP = true;
		}

		// percentage histogram
		if(stricmp(argv[i], "-p") == 0)
		{
			gPercentFlag = true;
		}

		// scaled histogram
		if(stricmp(argv[i], "-s") == 0)
		{
			gScaledFlag = true;
		}

		// sorted histogram
		if(stricmp(argv[i], "-sort") == 0)
		{
			gSortFlag = true;
		}

		// zoom factor - zooms in on lower part of histogram, useful when there is a large disparity
		// between the peak and most of the histogram
		if(stricmp(argv[i], "-z") == 0)
		{
			i++;
			gZoomFactor = atof(argv[i]);
			if(gZoomFactor < 1.0) gZoomFactor = 1.0;
			if(gZoomFactor > 100.0) gZoomFactor = 100.0;
		}

		// zoom to top part of histogram
		if(stricmp(argv[i], "-t") == 0)
		{
			gZoomTop = true;
		}
		i++;
	} // loop through command line arguments

	return(SUCCESS);
} // parseCommandLine

// main function
void main(int argc, char *argv[])
{
	int x;

	printf("\n\n Write Bitmap Histogram Version: %4.2f\n\n", VERSION);
	printf("    by John A. Ortiz, last update Jan 2012\n\n");

	x = parseCommandLine(argc, argv);
	if(x == FAILURE) return;

	initBM_Header();

	// makes a bitmap image of the data
	if(gCreateBMP) write_BMP(argv[1]); // if hWidth ==0 , then uses square root of input file size

	// makes a bitmap image of the histogram
	writeHistograms(argv[1]);


	return;
} // main
